"""Mixin class for exporting results."""

import base64
import csv
import io
import html

from typing import Literal, Optional, Union, List


class DatasetExportMixin:
    """Mixin class for exporting Dataset objects."""

    def relevant_columns(
        self, data_type: Optional[str] = None, remove_prefix=False
    ) -> list:
        """Return the set of keys that are present in the dataset.

        :param data_type: The data type to filter by.
        :param remove_prefix: Whether to remove the prefix from the column names.

        >>> from edsl.results.Dataset import Dataset
        >>> d = Dataset([{'a.b':[1,2,3,4]}])
        >>> d.relevant_columns()
        ['a.b']

        >>> d.relevant_columns(remove_prefix=True)
        ['b']

        >>> from edsl.results import Results; Results.example().select('how_feeling', 'how_feeling_yesterday').relevant_columns()
        ['answer.how_feeling', 'answer.how_feeling_yesterday']

        >>> from edsl.results import Results
        >>> sorted(Results.example().select().relevant_columns(data_type = "model"))
        ['model.frequency_penalty', 'model.logprobs', 'model.max_tokens', 'model.model', 'model.presence_penalty', 'model.temperature', 'model.top_logprobs', 'model.top_p']

        >>> Results.example().relevant_columns(data_type = "flimflam")
        Traceback (most recent call last):
        ...
        ValueError: No columns found for data type: flimflam. Available data types are: ['agent', 'answer', 'comment', 'model', 'prompt', 'question_options', 'question_text', 'question_type', 'raw_model_response', 'scenario'].
        """
        columns = [list(x.keys())[0] for x in self]
        if remove_prefix:
            columns = [column.split(".")[-1] for column in columns]

        def get_data_type(column):
            if "." in column:
                return column.split(".")[0]
            else:
                return None

        if data_type:
            all_columns = columns[:]
            columns = [
                column for column in columns if get_data_type(column) == data_type
            ]
            if len(columns) == 0:
                all_data_types = sorted(
                    list(set(get_data_type(column) for column in all_columns))
                )
                raise ValueError(
                    f"No columns found for data type: {data_type}. Available data types are: {all_data_types}."
                )

        return columns

    def num_observations(self):
        """Return the number of observations in the dataset.

        >>> from edsl.results import Results
        >>> Results.example().num_observations()
        4
        """
        _num_observations = None
        for entry in self:
            key, values = list(entry.items())[0]
            if _num_observations is None:
                _num_observations = len(values)
            else:
                if len(values) != _num_observations:
                    raise ValueError(
                        "The number of observations is not consistent across columns."
                    )

        return _num_observations

    def _make_tabular(
        self, remove_prefix: bool, pretty_labels: Optional[dict] = None
    ) -> tuple[list, List[list]]:
        """Turn the results into a tabular format.

        :param remove_prefix: Whether to remove the prefix from the column names.

        >>> from edsl.results import Results
        >>> r = Results.example()
        >>> r.select('how_feeling')._make_tabular(remove_prefix = True)
        (['how_feeling'], [['OK'], ['Great'], ['Terrible'], ['OK']])

        >>> r.select('how_feeling')._make_tabular(remove_prefix = True, pretty_labels = {'how_feeling': "How are you feeling"})
        (['How are you feeling'], [['OK'], ['Great'], ['Terrible'], ['OK']])
        """

        def create_dict_from_list_of_dicts(list_of_dicts):
            for entry in list_of_dicts:
                key, list_of_values = list(entry.items())[0]
                yield key, list_of_values

        tabular_repr = dict(create_dict_from_list_of_dicts(self.data))

        full_header = [list(x.keys())[0] for x in self]

        rows = []
        for i in range(self.num_observations()):
            row = [tabular_repr[h][i] for h in full_header]
            rows.append(row)

        if remove_prefix:
            header = [h.split(".")[-1] for h in full_header]
        else:
            header = full_header

        if pretty_labels is not None:
            header = [pretty_labels.get(h, h) for h in header]

        return header, rows

    def print_long(self):
        """Print the results in a long format.
        >>> from edsl.results import Results
        >>> r = Results.example()
        >>> r.select('how_feeling').print_long()
        answer.how_feeling: OK
        answer.how_feeling: Great
        answer.how_feeling: Terrible
        answer.how_feeling: OK
        """
        for entry in self:
            key, list_of_values = list(entry.items())[0]
            for value in list_of_values:
                print(f"{key}: {value}")

    def print(
        self,
        pretty_labels: Optional[dict] = None,
        filename: Optional[str] = None,
        format: Optional[Literal["rich", "html", "markdown", "latex"]] = None,
        interactive: bool = False,
        split_at_dot: bool = True,
        max_rows=None,
        tee=False,
        iframe=False,
        iframe_height: int = 200,
        iframe_width: int = 600,
        web=False,
    ) -> None:
        """Print the results in a pretty format.

        :param pretty_labels: A dictionary of pretty labels for the columns.
        :param filename: The filename to save the results to.
        :param format: The format to print the results in. Options are 'rich', 'html', or 'markdown'.
        :param interactive: Whether to print the results interactively in a Jupyter notebook.
        :param split_at_dot: Whether to split the column names at the last dot w/ a newline.
        :param max_rows: The maximum number of rows to print.
        :param tee: Whether to return the dataset.
        :param iframe: Whether to display the table in an iframe.
        :param iframe_height: The height of the iframe.
        :param iframe_width: The width of the iframe.
        :param web: Whether to display the table in a web browser.

        Example: Print in rich format at the terminal

        >>> from edsl.results import Results
        >>> r = Results.example()
        >>> r.select('how_feeling').print(format = "rich")
        ┏━━━━━━━━━━━━━━┓
        ┃ answer       ┃
        ┃ .how_feeling ┃
        ┡━━━━━━━━━━━━━━┩
        │ OK           │
        ├──────────────┤
        │ Great        │
        ├──────────────┤
        │ Terrible     │
        ├──────────────┤
        │ OK           │
        └──────────────┘

        >>> r = Results.example()
        >>> r2 = r.select("how_feeling").print(format = "rich", tee = True, max_rows = 2)
        ┏━━━━━━━━━━━━━━┓
        ┃ answer       ┃
        ┃ .how_feeling ┃
        ┡━━━━━━━━━━━━━━┩
        │ OK           │
        ├──────────────┤
        │ Great        │
        └──────────────┘
        >>> r2
        Dataset([{'answer.how_feeling': ['OK', 'Great', 'Terrible', 'OK']}])

        >>> r.select('how_feeling').print(format = "rich", max_rows = 2)
        ┏━━━━━━━━━━━━━━┓
        ┃ answer       ┃
        ┃ .how_feeling ┃
        ┡━━━━━━━━━━━━━━┩
        │ OK           │
        ├──────────────┤
        │ Great        │
        └──────────────┘

        >>> r.select('how_feeling').print(format = "rich", split_at_dot = False)
        ┏━━━━━━━━━━━━━━━━━━━━┓
        ┃ answer.how_feeling ┃
        ┡━━━━━━━━━━━━━━━━━━━━┩
        │ OK                 │
        ├────────────────────┤
        │ Great              │
        ├────────────────────┤
        │ Terrible           │
        ├────────────────────┤
        │ OK                 │
        └────────────────────┘

        Example: using the pretty_labels parameter

        >>> r.select('how_feeling').print(format="rich", pretty_labels = {'answer.how_feeling': "How are you feeling"})
        ┏━━━━━━━━━━━━━━━━━━━━━┓
        ┃ How are you feeling ┃
        ┡━━━━━━━━━━━━━━━━━━━━━┩
        │ OK                  │
        ├─────────────────────┤
        │ Great               │
        ├─────────────────────┤
        │ Terrible            │
        ├─────────────────────┤
        │ OK                  │
        └─────────────────────┘

        Example: printing in markdown format

        >>> r.select('how_feeling').print(format='markdown')
        | answer.how_feeling |
        |--|
        | OK |
        | Great |
        | Terrible |
        | OK |
        ...

        >>> r.select('how_feeling').print(format='latex')
        \\begin{tabular}{l}
        \\toprule
        ...
        """
        from IPython.display import HTML, display
        from edsl.utilities.utilities import is_notebook

        def _determine_format(format):
            if format is None:
                if is_notebook():
                    format = "html"
                else:
                    format = "rich"
            if format not in ["rich", "html", "markdown", "latex"]:
                raise ValueError("format must be one of 'rich', 'html', or 'markdown'.")

            return format

        format = _determine_format(format)

        if pretty_labels is None:
            pretty_labels = {}

        if pretty_labels != {}:  # only split at dot if there are no pretty labels
            split_at_dot = False

        def _create_data():
            for index, entry in enumerate(self):
                key, list_of_values = list(entry.items())[0]
                yield {pretty_labels.get(key, key): list_of_values[:max_rows]}

        new_data = list(_create_data())

        if format == "rich":
            from edsl.utilities.interface import print_dataset_with_rich

            print_dataset_with_rich(
                new_data, filename=filename, split_at_dot=split_at_dot
            )
            return self if tee else None

        if format == "markdown":
            from edsl.utilities.interface import print_list_of_dicts_as_markdown_table

            print_list_of_dicts_as_markdown_table(new_data, filename=filename)
            return self if tee else None

        if format == "latex":
            df = self.to_pandas()
            df.columns = [col.replace("_", " ") for col in df.columns]
            latex_string = df.to_latex(index=False)

            if filename is not None:
                with open(filename, "w") as f:
                    f.write(latex_string)
            else:
                print(latex_string)

            return self if tee else None

        if format == "html":
            from edsl.utilities.interface import print_list_of_dicts_as_html_table

            html_source = print_list_of_dicts_as_html_table(
                new_data, interactive=interactive
            )

            # if download_link:
            #     from IPython.display import HTML, display
            #     csv_file = output.getvalue()
            #     b64 = base64.b64encode(csv_file.encode()).decode()
            #     download_link = f'<a href="data:file/csv;base64,{b64}" download="my_data.csv">Download CSV file</a>'
            #     #display(HTML(download_link))

            if iframe:
                iframe = f""""
                <iframe srcdoc="{ html.escape(html_source) }" style="width: {iframe_width}px; height: {iframe_height}px;"></iframe>
                """
                display(HTML(iframe))
            elif is_notebook():
                display(HTML(html_source))
            else:
                from edsl.utilities.interface import view_html

                view_html(html_source)

            return self if tee else None

    def to_csv(
        self,
        filename: Optional[str] = None,
        remove_prefix: bool = False,
        download_link: bool = False,
        pretty_labels: Optional[dict] = None,
    ):
        """Export the results to a CSV file.

        :param filename: The filename to save the CSV file to.
        :param remove_prefix: Whether to remove the prefix from the column names.
        :param download_link: Whether to display a download link in a Jupyter notebook.

        Example:

        >>> from edsl.results import Results
        >>> r = Results.example()
        >>> r.select('how_feeling').to_csv()
        'answer.how_feeling\\r\\nOK\\r\\nGreat\\r\\nTerrible\\r\\nOK\\r\\n'

        >>> r.select('how_feeling').to_csv(pretty_labels = {'answer.how_feeling': "How are you feeling"})
        'How are you feeling\\r\\nOK\\r\\nGreat\\r\\nTerrible\\r\\nOK\\r\\n'

        >>> import tempfile
        >>> filename = tempfile.NamedTemporaryFile(delete=False).name
        >>> r.select('how_feeling').to_csv(filename = filename)
        >>> import os
        >>> import csv
        >>> with open(filename, newline='') as f:
        ...     reader = csv.reader(f)
        ...     for row in reader:
        ...         print(row)
        ['answer.how_feeling']
        ['OK']
        ['Great']
        ['Terrible']
        ['OK']

        """
        if pretty_labels is None:
            pretty_labels = {}
        header, rows = self._make_tabular(
            remove_prefix=remove_prefix, pretty_labels=pretty_labels
        )

        if filename is not None:
            with open(filename, "w") as f:
                writer = csv.writer(f)
                writer.writerow(header)
                writer.writerows(rows)
        else:
            output = io.StringIO()
            writer = csv.writer(output)
            writer.writerow(header)
            writer.writerows(rows)

            if download_link:
                from IPython.display import HTML, display

                csv_file = output.getvalue()
                b64 = base64.b64encode(csv_file.encode()).decode()
                download_link = f'<a href="data:file/csv;base64,{b64}" download="my_data.csv">Download CSV file</a>'
                display(HTML(download_link))
            else:
                return output.getvalue()

    def download_link(self, pretty_labels: Optional[dict] = None) -> str:
        """Return a download link for the results.

        :param pretty_labels: A dictionary of pretty labels for the columns.

        >>> from edsl.results import Results
        >>> r = Results.example()
        >>> r.select('how_feeling').download_link()
        '<a href="data:file/csv;base64,YW5zd2VyLmhvd19mZWVsaW5nDQpPSw0KR3JlYXQNClRlcnJpYmxlDQpPSw0K" download="my_data.csv">Download CSV file</a>'
        """
        import base64

        csv_string = self.to_csv(pretty_labels=pretty_labels)
        b64 = base64.b64encode(csv_string.encode()).decode()
        return f'<a href="data:file/csv;base64,{b64}" download="my_data.csv">Download CSV file</a>'

    def to_pandas(self, remove_prefix: bool = False) -> "pd.DataFrame":
        """Convert the results to a pandas DataFrame.

        :param remove_prefix: Whether to remove the prefix from the column names.

        >>> from edsl.results import Results
        >>> r = Results.example()
        >>> r.select('how_feeling').to_pandas()
          answer.how_feeling
        0                 OK
        1              Great
        2           Terrible
        3                 OK
        """
        import pandas as pd

        csv_string = self.to_csv(remove_prefix=remove_prefix)
        csv_buffer = io.StringIO(csv_string)
        df = pd.read_csv(csv_buffer)
        # df_sorted = df.sort_index(axis=1)  # Sort columns alphabetically
        return df

    def to_scenario_list(self, remove_prefix: bool = True) -> list[dict]:
        """Convert the results to a list of dictionaries, one per scenario.

        :param remove_prefix: Whether to remove the prefix from the column names.

        >>> from edsl.results import Results
        >>> r = Results.example()
        >>> r.select('how_feeling').to_scenario_list()
        ScenarioList([Scenario({'how_feeling': 'OK'}), Scenario({'how_feeling': 'Great'}), Scenario({'how_feeling': 'Terrible'}), Scenario({'how_feeling': 'OK'})])
        """
        from edsl import ScenarioList, Scenario

        list_of_dicts = self.to_dicts(remove_prefix=remove_prefix)
        return ScenarioList([Scenario(d) for d in list_of_dicts])

    def to_agent_list(self, remove_prefix: bool = True):
        """Convert the results to a list of dictionaries, one per agent.

        :param remove_prefix: Whether to remove the prefix from the column names.

        >>> from edsl.results import Results
        >>> r = Results.example()
        >>> r.select('how_feeling').to_agent_list()
        AgentList([Agent(traits = {'how_feeling': 'OK'}), Agent(traits = {'how_feeling': 'Great'}), Agent(traits = {'how_feeling': 'Terrible'}), Agent(traits = {'how_feeling': 'OK'})])
        """
        from edsl import AgentList, Agent

        list_of_dicts = self.to_dicts(remove_prefix=remove_prefix)
        return AgentList([Agent(d) for d in list_of_dicts])

    def to_dicts(self, remove_prefix: bool = True) -> list[dict]:
        """Convert the results to a list of dictionaries.

        :param remove_prefix: Whether to remove the prefix from the column names.

        >>> from edsl.results import Results
        >>> r = Results.example()
        >>> r.select('how_feeling').to_dicts()
        [{'how_feeling': 'OK'}, {'how_feeling': 'Great'}, {'how_feeling': 'Terrible'}, {'how_feeling': 'OK'}]

        """
        list_of_keys = []
        list_of_values = []
        for entry in self:
            key, values = list(entry.items())[0]
            list_of_keys.append(key)
            list_of_values.append(values)

        if remove_prefix:
            list_of_keys = [key.split(".")[-1] for key in list_of_keys]

        list_of_dicts = []
        for entries in zip(*list_of_values):
            list_of_dicts.append(dict(zip(list_of_keys, entries)))

        return list_of_dicts

    def to_list(self, flatten=False, remove_none=False) -> list[list]:
        """Convert the results to a list of lists.

        :param flatten: Whether to flatten the list of lists.
        :param remove_none: Whether to remove None values from the list.

        >>> from edsl.results import Results
        >>> Results.example().select('how_feeling', 'how_feeling_yesterday')
        Dataset([{'answer.how_feeling': ['OK', 'Great', 'Terrible', 'OK']}, {'answer.how_feeling_yesterday': ['Great', 'Good', 'OK', 'Terrible']}])

        >>> Results.example().select('how_feeling', 'how_feeling_yesterday').to_list()
        [('OK', 'Great'), ('Great', 'Good'), ('Terrible', 'OK'), ('OK', 'Terrible')]

        >>> r = Results.example()
        >>> r.select('how_feeling').to_list()
        ['OK', 'Great', 'Terrible', 'OK']

        >>> from edsl.results.Dataset import Dataset
        >>> Dataset([{'a.b': [[1, 9], 2, 3, 4]}]).select('a.b').to_list(flatten = True)
        [1, 9, 2, 3, 4]

        >>> from edsl.results.Dataset import Dataset
        >>> Dataset([{'a.b': [[1, 9], 2, 3, 4]}, {'c': [6, 2, 3, 4]}]).select('a.b', 'c').to_list(flatten = True)
        Traceback (most recent call last):
        ...
        ValueError: Cannot flatten a list of lists when there are multiple columns selected.


        """
        if len(self.relevant_columns()) > 1 and flatten:
            raise ValueError(
                "Cannot flatten a list of lists when there are multiple columns selected."
            )

        if len(self.relevant_columns()) == 1:
            # if only one 'column' is selected (which is typical for this method
            list_to_return = list(self[0].values())[0]
        else:
            keys = self.relevant_columns()
            data = self.to_dicts(remove_prefix=False)
            list_to_return = []
            for d in data:
                list_to_return.append(tuple([d[key] for key in keys]))

        if remove_none:
            list_to_return = [item for item in list_to_return if item is not None]

        if flatten:
            new_list = []
            for item in list_to_return:
                if isinstance(item, list):
                    new_list.extend(item)
                else:
                    new_list.append(item)
            list_to_return = new_list

        return list_to_return

    def html(
        self,
        filename: Optional[str] = None,
        cta: str = "Open in browser",
        return_link: bool = False,
    ):
        import os
        import tempfile
        from edsl.utilities.utilities import is_notebook
        from IPython.display import HTML, display
        from edsl.utilities.utilities import is_notebook

        df = self.to_pandas()

        if filename is None:
            current_directory = os.getcwd()
            filename = tempfile.NamedTemporaryFile(
                "w", delete=False, suffix=".html", dir=current_directory
            ).name

        with open(filename, "w") as f:
            f.write(df.to_html())

        if is_notebook():
            html_url = f"/files/{filename}"
            html_link = f'<a href="{html_url}" target="_blank">{cta}</a>'
            display(HTML(html_link))
        else:
            print(f"Saved to {filename}")
            import webbrowser
            import os

            webbrowser.open(f"file://{os.path.abspath(filename)}")

        if return_link:
            return filename

    def tally(
        self, *fields: Optional[str], top_n: Optional[int] = None, output="dict"
    ) -> Union[dict, "Dataset"]:
        """Tally the values of a field or perform a cross-tab of multiple fields.

        :param fields: The field(s) to tally, multiple fields for cross-tabulation.

        >>> from edsl.results import Results
        >>> r = Results.example()
        >>> r.select('how_feeling').tally('answer.how_feeling')
        {'OK': 2, 'Great': 1, 'Terrible': 1}
        >>> r.select('how_feeling', 'period').tally('how_feeling', 'period')
        {('OK', 'morning'): 1, ('Great', 'afternoon'): 1, ('Terrible', 'morning'): 1, ('OK', 'afternoon'): 1}
        """
        from collections import Counter

        if len(fields) == 0:
            fields = self.relevant_columns()

        relevant_columns_without_prefix = [
            column.split(".")[-1] for column in self.relevant_columns()
        ]

        if not all(
            f in self.relevant_columns() or f in relevant_columns_without_prefix
            for f in fields
        ):
            raise ValueError("One or more specified fields are not in the dataset.")

        if len(fields) == 1:
            field = fields[0]
            values = self._key_to_value(field)
        else:
            values = list(zip(*(self._key_to_value(field) for field in fields)))

        for value in values:
            if isinstance(value, list):
                value = tuple(value)

        tally = dict(Counter(values))
        sorted_tally = dict(sorted(tally.items(), key=lambda item: -item[1]))
        if top_n is not None:
            sorted_tally = dict(list(sorted_tally.items())[:top_n])

        import warnings
        import textwrap
        from edsl.results.Dataset import Dataset

        if output == "dict":
            warnings.warn(
                textwrap.dedent(
                    """\
                        The default output from tally will change to Dataset in the future.
                        Use output='Dataset' to get the Dataset object for now.
                        """
                )
            )
            return sorted_tally
        elif output == "Dataset":
            return Dataset(
                [
                    {"value": list(sorted_tally.keys())},
                    {"count": list(sorted_tally.values())},
                ]
            )


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



import requests


class CloudStorageManager:
    def __init__(self, secret_token=None):
        self.api_url = "https://bucket-server-tte53lsfxq-uc.a.run.app"
        self.secret_token = secret_token

    def get_signed_url(self, file_name, operation="upload"):
        """Get a signed URL for uploading or downloading a file."""

        if operation == "upload":
            if self.secret_token == None:
                raise "Set secret_token for upload permissions"
            headers = {
                "Authorization": self.secret_token,
                "Content-Type": "application/json",
            }
        else:
            headers = {
                "Content-Type": "application/json",
            }
        data = {"file_name": file_name}
        endpoint = f"{self.api_url}/generate-{operation}-signed-url"
        response = requests.post(endpoint, json=data, headers=headers)

        if response.status_code == 200:
            return response.json().get("signed_url")
        else:
            raise Exception(
                f"Failed to get signed URL: {response.status_code} {response.text}"
            )

    def upload_file(self, file_path, upload_file_name):
        """Upload a file to the signed URL."""
        signed_url = self.get_signed_url(upload_file_name, operation="upload")

        with open(file_path, "rb") as file:
            upload_response = requests.put(
                signed_url,
                data=file,
                headers={"Content-Type": "application/octet-stream"},
            )

            if upload_response.status_code == 200:
                print("File uploaded successfully")
            else:
                raise Exception(
                    f"Failed to upload file: {upload_response.status_code} {upload_response.text}"
                )

    def download_file(self, file_name, save_name):
        """Download a file from the signed URL."""

        signed_url = self.get_signed_url(file_name, operation="download")
        download_response = requests.get(signed_url, stream=True)

        if download_response.status_code == 200:
            with open(save_name, "wb") as file:
                for chunk in download_response.iter_content(chunk_size=8192):
                    file.write(chunk)
            print("File downloaded successfully")
        else:
            raise Exception(
                f"Failed to download file: {download_response.status_code} {download_response.text}"
            )

    def delete_file(self, file_name):
        """Delete a file from the cloud storage."""
        headers = {
            "Authorization": self.secret_token,
            "Content-Type": "application/json",
        }
        data = {"file_name": file_name}
        endpoint = f"{self.api_url}/delete-file"
        response = requests.delete(endpoint, params=data, headers=headers)

        if response.status_code == 200:
            print("File deleted successfully")
        else:
            raise Exception(
                f"Failed to delete file: {response.status_code} {response.text}"
            )

    def list_files(self):
        url = self.api_url + "/list_files"
        headers = {
            "Authorization": self.secret_token,
            "Content-Type": "application/json",
        }
        res = requests.get(url, headers=headers)
        data = res.json()
        for x in data["data"]:
            x["url"] = self.api_url + "/file/" + x["shaKey"]

        return data



"""This module contains the Interview class, which is responsible for conducting an interview asynchronously."""

from __future__ import annotations
import traceback
import asyncio
import time
from typing import Any, Type, List, Generator, Optional

from edsl.jobs.Answers import Answers
from edsl.surveys.base import EndOfSurvey
from edsl.jobs.buckets.ModelBuckets import ModelBuckets
from edsl.jobs.tasks.TaskCreators import TaskCreators

from edsl.jobs.interviews.InterviewStatusLog import InterviewStatusLog
from edsl.jobs.interviews.interview_exception_tracking import (
    InterviewExceptionCollection,
    InterviewExceptionEntry,
)
from edsl.jobs.interviews.retry_management import retry_strategy
from edsl.jobs.interviews.InterviewTaskBuildingMixin import InterviewTaskBuildingMixin
from edsl.jobs.interviews.InterviewStatusMixin import InterviewStatusMixin

import asyncio


def run_async(coro):
    return asyncio.run(coro)


class Interview(InterviewStatusMixin, InterviewTaskBuildingMixin):
    """
    An 'interview' is one agent answering one survey, with one language model, for a given scenario.

    The main method is `async_conduct_interview`, which conducts the interview asynchronously.
    """

    def __init__(
        self,
        agent: "Agent",
        survey: "Survey",
        scenario: "Scenario",
        model: Type["LanguageModel"],
        debug: Optional[bool] = False,
        iteration: int = 0,
        cache: Optional["Cache"] = None,
        sidecar_model: Optional["LanguageModel"] = None,
    ):
        """Initialize the Interview instance.

        :param agent: the agent being interviewed.
        :param survey: the survey being administered to the agent.
        :param scenario: the scenario that populates the survey questions.
        :param model: the language model used to answer the questions.
        :param debug: if True, run without calls to the language model.
        :param iteration: the iteration number of the interview.
        :param cache: the cache used to store the answers.
        :param sidecar_model: a sidecar model used to answer questions.

        >>> i = Interview.example()
        >>> i.task_creators
        {}

        >>> i.exceptions
        {}

        >>> _ = asyncio.run(i.async_conduct_interview())
        >>> i.task_status_logs['q0']
        [{'log_time': ..., 'value': <TaskStatus.NOT_STARTED: 1>}, {'log_time': ..., 'value': <TaskStatus.WAITING_FOR_DEPENDENCIES: 2>}, {'log_time': ..., 'value': <TaskStatus.API_CALL_IN_PROGRESS: 7>}, {'log_time': ..., 'value': <TaskStatus.SUCCESS: 8>}]

        >>> i.to_index
        {'q0': 0, 'q1': 1, 'q2': 2}

        """
        self.agent = agent
        self.survey = survey
        self.scenario = scenario
        self.model = model
        self.debug = debug
        self.iteration = iteration
        self.cache = cache
        self.answers: dict[
            str, str
        ] = Answers()  # will get filled in as interview progresses
        self.sidecar_model = sidecar_model

        # Trackers
        self.task_creators = TaskCreators()  # tracks the task creators
        self.exceptions = InterviewExceptionCollection()
        self._task_status_log_dict = InterviewStatusLog()

        # dictionary mapping question names to their index in the survey.
        self.to_index = {
            question_name: index
            for index, question_name in enumerate(self.survey.question_names)
        }

    async def async_conduct_interview(
        self,
        *,
        model_buckets: ModelBuckets = None,
        debug: bool = False,
        stop_on_exception: bool = False,
        sidecar_model: Optional["LanguageModel"] = None,
    ) -> tuple["Answers", List[dict[str, Any]]]:
        """
        Conduct an Interview asynchronously.
        It returns a tuple with the answers and a list of valid results.

        :param model_buckets: a dictionary of token buckets for the model.
        :param debug: run without calls to LLM.
        :param stop_on_exception: if True, stops the interview if an exception is raised.
        :param sidecar_model: a sidecar model used to answer questions.

        Example usage:

        >>> i = Interview.example()
        >>> result, _ = asyncio.run(i.async_conduct_interview())
        >>> result['q0']
        'yes'

        >>> i = Interview.example(throw_exception = True)
        >>> result, _ = asyncio.run(i.async_conduct_interview())
        Attempt 1 failed with exception:This is a test error now waiting 1.00 seconds before retrying.Parameters: start=1.0, max=60.0, max_attempts=5.
        <BLANKLINE>
        <BLANKLINE>
        Attempt 2 failed with exception:This is a test error now waiting 2.00 seconds before retrying.Parameters: start=1.0, max=60.0, max_attempts=5.
        <BLANKLINE>
        <BLANKLINE>
        Attempt 3 failed with exception:This is a test error now waiting 4.00 seconds before retrying.Parameters: start=1.0, max=60.0, max_attempts=5.
        <BLANKLINE>
        <BLANKLINE>
        Attempt 4 failed with exception:This is a test error now waiting 8.00 seconds before retrying.Parameters: start=1.0, max=60.0, max_attempts=5.
        <BLANKLINE>
        <BLANKLINE>

        >>> i.exceptions
        {'q0': [{'exception': "Exception('This is a test error')", 'time': ..., 'traceback': ...

        >>> i = Interview.example()
        >>> result, _ = asyncio.run(i.async_conduct_interview(stop_on_exception = True))
        Traceback (most recent call last):
        ...
        asyncio.exceptions.CancelledError
        """
        self.sidecar_model = sidecar_model

        # if no model bucket is passed, create an 'infinity' bucket with no rate limits
        if model_buckets is None or hasattr(self.agent, "answer_question_directly"):
            model_buckets = ModelBuckets.infinity_bucket()

        ## build the tasks using the InterviewTaskBuildingMixin
        ## This is the key part---it creates a task for each question,
        ## with dependencies on the questions that must be answered before this one can be answered.
        self.tasks = self._build_question_tasks(
            debug=debug, model_buckets=model_buckets
        )

        ## 'Invigilators' are used to administer the survey
        self.invigilators = list(self._build_invigilators(debug=debug))
        # await the tasks being conducted
        await asyncio.gather(*self.tasks, return_exceptions=not stop_on_exception)
        self.answers.replace_missing_answers_with_none(self.survey)
        valid_results = list(self._extract_valid_results())
        return self.answers, valid_results

    def _extract_valid_results(self) -> Generator["Answers", None, None]:
        """Extract the valid results from the list of results.

        It iterates through the tasks and invigilators, and yields the results of the tasks that are done.
        If a task is not done, it raises a ValueError.
        If an exception is raised in the task, it records the exception in the Interview instance except if the task was cancelled, which is expected behavior.

        >>> i = Interview.example()
        >>> result, _ = asyncio.run(i.async_conduct_interview())
        >>> results = list(i._extract_valid_results())
        >>> len(results) == len(i.survey)
        True
        >>> type(results[0])
        <class 'edsl.data_transfer_models.AgentResponseDict'>
        """
        assert len(self.tasks) == len(self.invigilators)

        for task, invigilator in zip(self.tasks, self.invigilators):
            if not task.done():
                raise ValueError(f"Task {task.get_name()} is not done.")

            try:
                result = task.result()
            except asyncio.CancelledError as e:  # task was cancelled
                result = invigilator.get_failed_task_result()
            except Exception as e:  # any other kind of exception in the task
                result = invigilator.get_failed_task_result()
                self._record_exception(task, e)
            yield result

    def _record_exception(self, task, exception: Exception) -> None:
        """Record an exception in the Interview instance.

        It records the exception in the Interview instance, with the task name and the exception entry.

        >>> i = Interview.example()
        >>> result, _ = asyncio.run(i.async_conduct_interview())
        >>> i.exceptions
        {}
        >>> i._record_exception(i.tasks[0], Exception("An exception occurred."))
        >>> i.exceptions
        {'q0': [{'exception': "Exception('An exception occurred.')", 'time': ..., 'traceback': 'NoneType: None\\n'}]}
        """
        exception_entry = InterviewExceptionEntry(
            exception=repr(exception),
            time=time.time(),
            traceback=traceback.format_exc(),
        )
        self.exceptions.add(task.get_name(), exception_entry)

    @property
    def dag(self) -> "DAG":
        """Return the directed acyclic graph for the survey.

        The DAG, or directed acyclic graph, is a dictionary that maps question names to their dependencies.
        It is used to determine the order in which questions should be answered.
        This reflects both agent 'memory' considerations and 'skip' logic.
        The 'textify' parameter is set to True, so that the question names are returned as strings rather than integer indices.

        >>> i = Interview.example()
        >>> i.dag == {'q2': {'q0'}, 'q1': {'q0'}}
        True
        """
        return self.survey.dag(textify=True)

    #######################
    # Dunder methods
    #######################
    def __repr__(self) -> str:
        """Return a string representation of the Interview instance."""
        return f"Interview(agent = {repr(self.agent)}, survey = {repr(self.survey)}, scenario = {repr(self.scenario)}, model = {repr(self.model)})"

    def duplicate(self, iteration: int, cache: "Cache") -> Interview:
        """Duplicate the interview, but with a new iteration number and cache.

        >>> i = Interview.example()
        >>> i2 = i.duplicate(1, None)
        >>> i.iteration + 1 == i2.iteration
        True

        """
        return Interview(
            agent=self.agent,
            survey=self.survey,
            scenario=self.scenario,
            model=self.model,
            iteration=iteration,
            cache=cache,
        )

    @classmethod
    def example(self, throw_exception: bool = False) -> Interview:
        """Return an example Interview instance."""
        from edsl.agents import Agent
        from edsl.surveys import Survey
        from edsl.scenarios import Scenario
        from edsl.language_models import LanguageModel

        def f(self, question, scenario):
            return "yes"

        agent = Agent.example()
        agent.add_direct_question_answering_method(f)
        survey = Survey.example()
        scenario = Scenario.example()
        model = LanguageModel.example()
        if throw_exception:
            model = LanguageModel.example(test_model=True, throw_exception=True)
            agent = Agent.example()
            return Interview(agent=agent, survey=survey, scenario=scenario, model=model)
        return Interview(agent=agent, survey=survey, scenario=scenario, model=model)


if __name__ == "__main__":
    import doctest

    # add ellipsis
    doctest.testmod(optionflags=doctest.ELLIPSIS)



from __future__ import annotations
import textwrap
from random import uniform
from typing import Any, Optional, Union

from edsl.exceptions import QuestionAnswerValidationError
from edsl.questions.QuestionBase import QuestionBase
from edsl.questions.descriptors import NumericalOrNoneDescriptor


class QuestionNumerical(QuestionBase):
    """This question prompts the agent to answer with a numerical value."""

    question_type = "numerical"
    min_value: Optional[float] = NumericalOrNoneDescriptor()
    max_value: Optional[float] = NumericalOrNoneDescriptor()

    def __init__(
        self,
        question_name: str,
        question_text: str,
        min_value: Optional[Union[int, float]] = None,
        max_value: Optional[Union[int, float]] = None,
    ):
        """Initialize the question.

        :param question_name: The name of the question.
        :param question_text: The text of the question.
        :param instructions: Instructions for the question. If not provided, the default instructions are used. To view them, run `QuestionNumerical.default_instructions`.
        :param min_value: The minimum value of the answer.
        :param max_value: The maximum value of the answer.
        """
        self.question_name = question_name
        self.question_text = question_text
        self.min_value = min_value
        self.max_value = max_value

    ################
    # Answer methods
    ################
    def _validate_answer(
        self, answer: dict[str, Any]
    ) -> dict[str, Union[str, float, int]]:
        """Validate the answer."""
        self._validate_answer_template_basic(answer)
        self._validate_answer_key_value_numeric(answer, "answer")
        self._validate_answer_numerical(answer)
        return answer

    def _translate_answer_code_to_answer(self, answer, scenario: "Scenario" = None):
        """There is no answer code."""
        return answer

    def _simulate_answer(self, human_readable: bool = True):
        """Simulate a valid answer for debugging purposes."""
        from edsl.utilities.utilities import random_string

        return {
            "answer": uniform(self.min_value, self.max_value),
            "comment": random_string(),
        }

    @property
    def question_html_content(self) -> str:
        from jinja2 import Template

        question_html_content = Template(
            """
        <div>
        <input type="number" id="{{ question_name }}" name="{{ question_name }}">
        </div>
        """
        ).render(question_name=self.question_name)
        return question_html_content

    ################
    # Helpful methods
    ################
    @classmethod
    def example(cls) -> QuestionNumerical:
        """Return an example question."""
        return cls(
            question_name="age",
            question_text="How old are you in years?",
            min_value=0,
            max_value=86.7,
        )


def main():
    """Show example usage."""
    from edsl.questions.QuestionNumerical import QuestionNumerical

    q = QuestionNumerical.example()
    q.question_text
    q.min_value
    q.max_value
    # validate an answer
    q._validate_answer({"answer": 1, "comment": "I like custard"})
    # translate answer code
    q._translate_answer_code_to_answer(1)
    # simulate answer
    q._simulate_answer()
    q._simulate_answer(human_readable=False)
    q._validate_answer(q._simulate_answer(human_readable=False))
    # serialization (inherits from Question)
    q.to_dict()
    assert q.from_dict(q.to_dict()) == q

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



import random
from typing import Generator, List, Optional, Union, Callable
from edsl.agents.Agent import Agent
from edsl.agents.AgentList import AgentList
from edsl.questions import QuestionBase
from edsl.results.Results import Results


class AgentConstructionMixin:
    def agent(self, index) -> Agent:
        """Return an agent constructed from the data.

        :param index: The index of the agent to construct.

        >>> from edsl.conjure.InputData import InputDataABC
        >>> id = InputDataABC.example()
        >>> id.agent(0)
        Agent(traits = {'morning': '1', 'feeling': '3'}, codebook = {'morning': 'how are you doing this morning?', 'feeling': 'how are you feeling?'})


        """
        responses = [responses[index] for responses in self.raw_data]
        traits = {qn: r for qn, r in zip(self.question_names, responses)}

        a = Agent(traits=traits, codebook=self.names_to_texts)

        def construct_answer_dict_function(traits: dict) -> Callable:
            def func(self, question: "QuestionBase", scenario=None):
                return traits.get(question.question_name, None)

            return func

        a.add_direct_question_answering_method(construct_answer_dict_function(traits))
        return a

    def _agents(self, indices) -> Generator[Agent, None, None]:
        """Return a generator of agents, one for each index."""
        for idx in indices:
            yield self.agent(idx)

    def to_agent_list(
        self,
        indices: Optional[List] = None,
        sample_size: int = None,
        seed: str = "edsl",
        remove_direct_question_answering_method: bool = True,
    ) -> AgentList:
        """Return an AgentList from the data.

        :param indices: The indices of the agents to include.
        :param sample_size: The number of agents to sample.
        :param seed: The seed for the random number generator.

        >>> from edsl.conjure.InputData import InputDataABC
        >>> id = InputDataABC.example()
        >>> al = id.to_agent_list()
        >>> len(al) == id.num_observations
        True
        >>> al = id.to_agent_list(indices = [0, 1, 2])
        Traceback (most recent call last):
        ...
        ValueError: Index 2 is greater than the number of agents 2.
        """
        if indices and (sample_size or seed != "edsl"):
            raise ValueError(
                "You cannot pass both indices and sample_size/seed, as these are mutually exclusive."
            )

        if indices:
            if len(indices) == 0:
                raise ValueError("Indices must be a non-empty list.")
            if max(indices) >= self.num_observations:
                raise ValueError(
                    f"Index {max(indices)} is greater than the number of agents {self.num_observations}."
                )
            if min(indices) < 0:
                raise ValueError(f"Index {min(indices)} is less than 0.")

        if indices is None:
            if sample_size is None:
                indices = range(self.num_observations)
            else:
                if sample_size > self.num_observations:
                    raise ValueError(
                        f"Sample size {sample_size} is greater than the number of agents {self.num_observations}."
                    )
                random.seed(seed)
                indices = random.sample(range(self.num_observations), sample_size)

        agents = list(self._agents(indices))
        if remove_direct_question_answering_method:
            for a in agents:
                a.remove_direct_question_answering_method()
        return AgentList(agents)

    def to_results(
        self,
        indices: Optional[List] = None,
        sample_size: int = None,
        seed: str = "edsl",
        dryrun=False,
    ) -> Union[Results, None]:
        """Return the results of the survey.

        :param indices: The indices of the agents to include.
        :param sample_size: The number of agents to sample.
        :param seed: The seed for the random number generator.
        :param dryrun: If True, the survey will not be run, but the time to run it will be printed.

        >>> from edsl.conjure.InputData import InputDataABC
        >>> id = InputDataABC.example()
        >>> r = id.to_results()
        >>> len(r) == id.num_observations
        True
        """
        agent_list = self.to_agent_list(
            indices=indices,
            sample_size=sample_size,
            seed=seed,
            remove_direct_question_answering_method=False,
        )
        DRYRUN_SAMPLE = 30
        survey = self.to_survey()
        if dryrun:
            import time

            start = time.time()
            _ = survey.by(agent_list.sample(DRYRUN_SAMPLE)).run()
            end = time.time()
            print(f"Time to run {DRYRUN_SAMPLE} agents (s): {round(end - start, 2)}")
            time_per_agent = (end - start) / DRYRUN_SAMPLE
            full_sample_time = time_per_agent * len(agent_list)
            if full_sample_time < 60:
                print(
                    f"Full sample will take about {round(full_sample_time, 2)} seconds."
                )
            if full_sample_time > 60 and full_sample_time < 3600:
                print(
                    f"Full sample will take about {round(full_sample_time / 60, 2)} minutes."
                )
            if full_sample_time > 3600:
                print(
                    f"Full sample will take about {round(full_sample_time / 3600, 2)} hours."
                )
            return None
        return survey.by(agent_list).run()


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from collections import UserList


class TaskStatusLog(UserList):
    """A list of TaskStatusEntry objects."""

    @property
    def min_time(self):
        return self[0]["log_time"]

    @property
    def max_time(self):
        return self[-1]["log_time"]

    def status_at_time(self, t):
        """Return the status at time t.

        TODO: Could re-factor with bisect to make this faster.
        """
        for entry in self:
            if entry["log_time"] > t:
                return entry["value"]
        return self[-1]["value"]



"""Linear scale question type."""

import textwrap

from edsl.prompts.QuestionInstructionsBase import QuestionInstuctionsBase


class LinearScale(QuestionInstuctionsBase):
    """Linear scale question type."""

    question_type = "linear_scale"
    model = "gpt-4-1106-preview"
    default_instructions = textwrap.dedent(
        """\
        You are being asked the following question: {{question_text}}
        The options are
        {% for option in question_options %}
        {{ loop.index0 }}: {{option}}
        {% endfor %}
        Return a valid JSON formatted like this, selecting only the code of the option (codes start at 0):
        {"answer": <put answer code here>, "comment": "<put explanation here>"}
        Only 1 option may be selected.
        """
    )



import pandas as pd
from typing import List

from edsl.conjure.InputData import InputDataABC
from edsl.conjure.utilities import convert_value
from edsl.utilities.utilities import is_valid_variable_name

try:
    import pyreadstat
except ImportError as e:
    raise ImportError(
        "The 'pyreadstat' package is required for this feature. Please install it by running:\n"
        "pip install pyreadstat\n"
    ) from e


class InputDataPyRead(InputDataABC):
    def pyread_function(self, datafile_name):
        raise NotImplementedError

    def _parse(self) -> None:
        try:
            df, meta = self.pyread_function(self.datafile_name)
        except Exception as e:
            raise ValueError(
                f"An error occurred while reading the file {self.datafile_name}."
            ) from e
        float_columns = df.select_dtypes(include=["float64"]).columns
        df[float_columns] = df[float_columns].astype(str)

        df.fillna("", inplace=True)
        df = df.astype(str)
        self._df = df
        self._meta = meta

    def get_df(self) -> pd.DataFrame:
        if not hasattr(self, "_df"):
            self._parse()
        return self._df

    def get_answer_codebook(self):
        if not hasattr(self, "_meta"):
            self._parse()

        question_name_to_label_name = self._meta.variable_to_label
        label_name_to_labels = self._meta.value_labels
        return {
            qn: label_name_to_labels[label_name]
            for qn, label_name in question_name_to_label_name.items()
        }

    def get_raw_data(self) -> List[List[str]]:
        df = self.get_df()
        data = [
            [convert_value(obs) for obs in v]
            for k, v in df.to_dict(orient="list").items()
        ]
        return data

    @property
    def question_names_to_question_texts(self):
        """Return a dictionary of question names to question texts.
        This will repair the question names if they are not valid Python identifiers using the
        same question_name_repair_func that was passed in.
        """
        if not hasattr(self, "_meta"):
            self._parse()
        d = {}
        for qn, label in self._meta.column_names_to_labels.items():
            new_name = qn
            if not is_valid_variable_name(qn):
                new_name = self.question_name_repair_func(qn)
                if not is_valid_variable_name(new_name):
                    raise ValueError(
                        f"""Question names must be valid Python identifiers. '{qn}' is not.""",
                        """You can pass an entry in question_name_repair_dict to fix this.""",
                    )
            if label is not None:
                d[new_name] = label
        return d

    def get_question_texts(self):
        if not hasattr(self, "_meta"):
            self._parse()
        return [
            self.question_names_to_question_texts.get(qn, qn)
            for qn in self.question_names
        ]

    def get_question_names(self):
        return self.get_df().columns.tolist()



import base64
import os
import requests
import tempfile
import mimetypes
from urllib.parse import urlparse


class ScenarioImageMixin:
    def add_image(self, image_path: str):
        """Add an image to a scenario.

        >>> from edsl.scenarios.Scenario import Scenario
        >>> s = Scenario({"food": "wood chips"})
        >>> s.add_image(Scenario.example_image())
        Scenario({'food': 'wood chips', 'file_path': '...', 'encoded_image': '...'})
        """
        new_scenario = self.from_image(image_path)
        return self + new_scenario

    @staticmethod
    def example_image():
        """Return an example image path."""
        import os

        base_path = os.path.dirname(os.path.abspath(__file__))
        return os.path.join(base_path, "../../static/logo.png")

    @classmethod
    def from_image(cls, image_path: str) -> "Scenario":
        """Creates a scenario with a base64 encoding of an image.

        >>> from edsl.scenarios.Scenario import Scenario
        >>> s = Scenario.from_image(Scenario.example_image())
        >>> s
        Scenario({'file_path': '...', 'encoded_image': '...'})
        """

        if image_path.startswith("http://") or image_path.startswith("https://"):
            return cls._from_url_image(image_path)
        else:
            return cls._from_filepath_image(image_path)

    @classmethod
    def _from_url_image(cls, image_url: str) -> "Scenario":
        """Handles downloading and encoding an image from a URL."""
        response = requests.get(image_url)
        if response.status_code == 200:
            # Try to extract the file extension from the URL
            parsed_url = urlparse(image_url)
            file_name = parsed_url.path.split("/")[-1]
            file_extension = file_name.split(".")[-1] if "." in file_name else None

            # If the file extension is not found in the URL, use the content type
            if not file_extension:
                content_type = response.headers.get("Content-Type")
                file_extension = mimetypes.guess_extension(content_type)

            # If still no file extension, use a generic binary extension
            if not file_extension:
                file_extension = ".bin"

            # Create a temporary file with the appropriate extension
            with tempfile.NamedTemporaryFile(
                delete=False, suffix=file_extension
            ) as temp_file:
                # Write the image content to the temporary file
                temp_file.write(response.content)
                temp_file_name = temp_file.name
        else:
            raise ValueError("Failed to download the image.")

        scenario = cls._from_filepath_image(temp_file_name)
        os.remove(temp_file_name)
        return scenario

    @classmethod
    def _from_filepath_image(cls, image_path: str) -> "Scenario":
        """Handles encoding an image from a local file path."""
        with open(image_path, "rb") as image_file:
            s = cls(
                {
                    "file_path": image_path,
                    "encoded_image": base64.b64encode(image_file.read()).decode(
                        "utf-8"
                    ),
                }
            )
            s._has_image = True
            return s

    def __repr__(self):
        return f"Scenario({self.data})"


if __name__ == "__main__":
    import doctest
    from edsl.scenarios.Scenario import Scenario

    doctest.testmod(extraglobs={"Scenario": Scenario}, optionflags=doctest.ELLIPSIS)



"""A module to represent a dataset of observations."""

from __future__ import annotations
import random
from collections import UserList
from typing import Any, Union, Optional

import numpy as np

from edsl.results.ResultsExportMixin import ResultsExportMixin


class Dataset(UserList, ResultsExportMixin):
    """A class to represent a dataset of observations."""

    def __init__(self, data: list[dict[str, Any]] = None):
        """Initialize the dataset with the given data."""
        super().__init__(data)

    def __len__(self) -> int:
        """Return the number of observations in the dataset.

        Need to override the __len__ method to return the number of observations in the dataset because
        otherwise, the UserList class would return the number of dictionaries in the dataset.

        >>> d = Dataset([{'a.b':[1,2,3,4]}])
        >>> len(d)
        4
        """
        _, values = list(self.data[0].items())[0]
        return len(values)

    def __repr__(self) -> str:
        """Return a string representation of the dataset."""
        return f"Dataset({self.data})"

    def _key_to_value(self, key: str) -> Any:
        """Retrieve the value associated with the given key from the dataset.

        >>> d = Dataset([{'a.b':[1,2,3,4]}])
        >>> d._key_to_value('a.b')
        [1, 2, 3, 4]

        >>> d._key_to_value('a')
        Traceback (most recent call last):
        ...
        KeyError: "Key 'a' not found in any of the dictionaries."

        """
        potential_matches = []
        for data_dict in self.data:
            data_key, data_values = list(data_dict.items())[0]
            if key == data_key:
                return data_values
            if key == data_key.split(".")[-1]:
                potential_matches.append((data_key, data_values))

        if len(potential_matches) == 1:
            return potential_matches[0][1]
        elif len(potential_matches) > 1:
            raise KeyError(
                f"Key '{key}' found in more than one location: {[m[0] for m in potential_matches]}"
            )

        raise KeyError(f"Key '{key}' not found in any of the dictionaries.")

    def first(self) -> dict[str, Any]:
        """Get the first value of the first key in the first dictionary.

        >>> d = Dataset([{'a.b':[1,2,3,4]}])
        >>> d.first()
        1
        """

        def get_values(d):
            """Get the values of the first key in the dictionary."""
            return list(d.values())[0]

        return get_values(self.data[0])[0]

    def select(self, *keys):
        """Return a new dataset with only the selected keys.

        :param keys: The keys to select.

        >>> d = Dataset([{'a.b':[1,2,3,4]}, {'c.d':[5,6,7,8]}])
        >>> d.select('a.b')
        Dataset([{'a.b': [1, 2, 3, 4]}])

        >>> d.select('a.b', 'c.d')
        Dataset([{'a.b': [1, 2, 3, 4]}, {'c.d': [5, 6, 7, 8]}])
        """
        if isinstance(keys, str):
            keys = [keys]

        new_data = []
        for observation in self.data:
            observation_key = list(observation.keys())[0]
            if observation_key in keys:
                new_data.append(observation)
        return Dataset(new_data)

    def _repr_html_(self) -> str:
        """Return an HTML representation of the dataset."""
        from edsl.utilities.utilities import data_to_html

        return data_to_html(self.data)

    def shuffle(self, seed=None) -> Dataset:
        """Return a new dataset with the observations shuffled.

        >>> d = Dataset([{'a.b':[1,2,3,4]}])
        >>> d.shuffle(seed=0)
        Dataset([{'a.b': [3, 1, 2, 4]}])
        """
        if seed is not None:
            random.seed(seed)

        indices = None

        for entry in self:
            key, values = list(entry.items())[0]
            if indices is None:
                indices = list(range(len(values)))
                random.shuffle(indices)
            entry[key] = [values[i] for i in indices]

        return self

    def sample(
        self,
        n: int = None,
        frac: float = None,
        with_replacement: bool = True,
        seed: Union[str, int, float] = None,
    ) -> Dataset:
        """Return a new dataset with a sample of the observations.

        :param n: The number of samples to take.
        :param frac: The fraction of samples to take.
        :param with_replacement: Whether to sample with replacement.
        :param seed: The seed for the random number generator.

        >>> d = Dataset([{'a.b':[1,2,3,4]}])
        >>> d.sample(n=2, seed=0, with_replacement=True)
        Dataset([{'a.b': [4, 4]}])

        >>> d.sample(n = 10, seed=0, with_replacement=False)
        Traceback (most recent call last):
        ...
        ValueError: Sample size cannot be greater than the number of available elements when sampling without replacement.
        """
        if seed is not None:
            random.seed(seed)

        # Validate the input for sampling parameters
        if n is None and frac is None:
            raise ValueError("Either 'n' or 'frac' must be provided for sampling.")

        if n is not None and frac is not None:
            raise ValueError("Only one of 'n' or 'frac' should be specified.")

        # Get the length of the lists from the first entry
        first_key, first_values = list(self[0].items())[0]
        total_length = len(first_values)

        # Determine the number of samples based on 'n' or 'frac'
        if n is None:
            n = int(total_length * frac)

        if not with_replacement and n > total_length:
            raise ValueError(
                "Sample size cannot be greater than the number of available elements when sampling without replacement."
            )

        # Sample indices based on the method chosen
        if with_replacement:
            indices = [random.randint(0, total_length - 1) for _ in range(n)]
        else:
            indices = random.sample(range(total_length), k=n)

        # Apply the same indices to all entries
        for entry in self:
            key, values = list(entry.items())[0]
            entry[key] = [values[i] for i in indices]

        return self

    def order_by(self, sort_key: str, reverse: bool = False) -> Dataset:
        """Return a new dataset with the observations sorted by the given key.

        :param sort_key: The key to sort the observations by.
        :param reverse: Whether to sort in reverse order.

        >>> d = Dataset([{'a':[1,2,3,4]}, {'b':[4,3,2,1]}])
        >>> d.order_by('a')
        Dataset([{'a': [1, 2, 3, 4]}, {'b': [4, 3, 2, 1]}])

        >>> d.order_by('a', reverse=True)
        Dataset([{'a': [4, 3, 2, 1]}, {'b': [1, 2, 3, 4]}])

        >>> d = Dataset([{'X.a':[1,2,3,4]}, {'X.b':[4,3,2,1]}])
        >>> d.order_by('a')
        Dataset([{'X.a': [1, 2, 3, 4]}, {'X.b': [4, 3, 2, 1]}])


        """

        def sort_indices(lst: list[Any]) -> list[int]:
            """
            Return the indices that would sort the list.

            :param lst: The list to be sorted.
            :return: A list of indices that would sort the list.
            """
            indices = np.argsort(lst).tolist()
            if reverse:
                indices.reverse()
            return indices

        number_found = 0
        for obs in self.data:
            key, values = list(obs.items())[0]
            # an obseration is {'a':[1,2,3,4]}
            # key = list(obs.keys())[0]
            if (
                sort_key == key or sort_key == key.split(".")[-1]
            ):  # e.g., "age" in "scenario.age"
                relevant_values = values
                number_found += 1

        if number_found == 0:
            raise ValueError(f"Key '{sort_key}' not found in any of the dictionaries.")
        elif number_found > 1:
            raise ValueError(f"Key '{sort_key}' found in more than one dictionary.")

        # relevant_values = self._key_to_value(sort_key)
        sort_indices_list = sort_indices(relevant_values)
        new_data = []
        for observation in self.data:
            # print(observation)
            key, values = list(observation.items())[0]
            new_values = [values[i] for i in sort_indices_list]
            new_data.append({key: new_values})

        return Dataset(new_data)

    @classmethod
    def example(self):
        """Return an example dataset.

        >>> Dataset.example()
        Dataset([{'a': [1, 2, 3, 4]}, {'b': [4, 3, 2, 1]}])
        """
        return Dataset([{"a": [1, 2, 3, 4]}, {"b": [4, 3, 2, 1]}])


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



"""Mixin class for exporting results."""

from functools import wraps
from edsl.results.DatasetExportMixin import DatasetExportMixin


def to_dataset(func):
    """Convert the object to a Dataset object before calling the function."""

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        """Return the function with the Results object converted to a Dataset object."""
        if self.__class__.__name__ == "ScenarioList":
            return func(self.to_dataset(), *args, **kwargs)
        else:
            raise Exception(
                f"Class {self.__class__.__name__} not recognized as a Results or Dataset object."
            )

    return wrapper


def decorate_all_methods(cls):
    for attr_name, attr_value in cls.__dict__.items():
        if callable(attr_value):
            setattr(cls, attr_name, to_dataset(attr_value))
    return cls


@decorate_all_methods
class ScenarioListExportMixin(DatasetExportMixin):
    """Mixin class for exporting Results objects."""



"""Agent instructions for a human agent."""

import textwrap

from edsl.prompts.Prompt import PromptBase
from edsl.prompts.prompt_config import ComponentTypes

# from edsl.enums import LanguageModelType


class AgentInstruction(PromptBase):
    """Agent instructions for a human agent."""

    # model = LanguageModelType.GPT_3_5_Turbo.value
    # from edsl import Model
    # model = Model().model
    model = "gpt-3.5-turbo"
    component_type = ComponentTypes.AGENT_INSTRUCTIONS
    default_instructions = textwrap.dedent(
        """\
    You are playing the role of a human answering survey questions.
    Do not break character.
    """
    )


# class AgentInstructionLlama(PromptBase):
#     """Agent instructions for a human agent."""

#     model = LanguageModelType.LLAMA_2_70B_CHAT_HF.value
#     component_type = ComponentTypes.AGENT_INSTRUCTIONS
#     default_instructions = textwrap.dedent(
#         """\
#     You are playing the role of a human answering questions.
#     Do not break character.
#     Only respond in JSON, with one answer formatted as specified.
#     """
#     )



from rich.console import Console
from rich.table import Table
from collections import UserDict


class InterviewExceptionEntry(UserDict):
    """Class to record an exception that occurred during the interview."""

    def __init__(self, exception, time, traceback):
        data = {"exception": exception, "time": time, "traceback": traceback}
        super().__init__(data)

    def to_dict(self) -> dict:
        """Return the exception as a dictionary."""
        return self.data


class InterviewExceptionCollection(UserDict):
    """A collection of exceptions that occurred during the interview."""

    def add(self, question_name: str, entry: InterviewExceptionEntry) -> None:
        """Add an exception entry to the collection."""
        question_name = question_name
        if question_name not in self.data:
            self.data[question_name] = []
        self.data[question_name].append(entry)

    def to_dict(self, include_traceback=True) -> dict:
        """Return the collection of exceptions as a dictionary."""
        newdata = {k: [e.to_dict() for e in v] for k, v in self.data.items()}
        # if not include_traceback:
        #     for question in newdata:
        #         for exception in newdata[question]:
        #             exception[
        #                 "traceback"
        #             ] = "Traceback removed. Set include_traceback=True to include."
        return newdata

    def _repr_html_(self) -> str:
        from edsl.utilities.utilities import data_to_html

        return data_to_html(self.to_dict(include_traceback=True))

    def ascii_table(self, traceback: bool = False) -> None:
        headers = ["Question name", "Exception", "Time", "Traceback"]
        from tabulate import tabulate

        data = []
        for question, exceptions in self.data.items():
            for exception in exceptions:
                if traceback:
                    row = [
                        question,
                        exception["exception"],
                        exception["time"],
                        exception["traceback"],
                    ]
                else:
                    row = [question, exception["exception"], exception["time"]]
                data.append(row)

        print(tabulate(data, headers=headers, tablefmt="grid"))

    def print(self, traceback=False):
        """Print the collection of exceptions."""
        console = Console()
        table = Table(show_header=True, header_style="bold magenta")
        table.add_column("Question name", style="dim", width=12)
        table.add_column("Exception", width=32)
        table.add_column("Time", justify="right")
        table.add_column("Traceback", min_width=20)

        for queue, exceptions in self.data.items():
            for exception in exceptions:
                if traceback:
                    traceback_string = exception["traceback"].replace("\n", "\n\n")
                else:
                    traceback_string = ""
                table.add_row(
                    queue,
                    exception["exception"],
                    str(exception["time"]),
                    traceback_string,  # Adding extra newlines for better readability
                )

        console.print(table)



import os
from typing import Any
import re
from anthropic import AsyncAnthropic
from edsl.inference_services.InferenceServiceABC import InferenceServiceABC
from edsl.language_models.LanguageModel import LanguageModel


class AnthropicService(InferenceServiceABC):
    """Anthropic service class."""

    _inference_service_ = "anthropic"
    _env_key_name_ = "ANTHROPIC_API_KEY"

    @classmethod
    def available(cls):
        # TODO - replace with an API call
        return [
            "claude-3-5-sonnet-20240620",
            "claude-3-opus-20240229",
            "claude-3-sonnet-20240229",
            "claude-3-haiku-20240307",
        ]

    @classmethod
    def create_model(
        cls, model_name: str = "claude-3-opus-20240229", model_class_name=None
    ) -> LanguageModel:
        if model_class_name is None:
            model_class_name = cls.to_class_name(model_name)

        class LLM(LanguageModel):
            """
            Child class of LanguageModel for interacting with OpenAI models
            """

            _inference_service_ = cls._inference_service_
            _model_ = model_name
            _parameters_ = {
                "temperature": 0.5,
                "max_tokens": 1000,
                "top_p": 1,
                "frequency_penalty": 0,
                "presence_penalty": 0,
                "logprobs": False,
                "top_logprobs": 3,
            }

            async def async_execute_model_call(
                self, user_prompt: str, system_prompt: str = ""
            ) -> dict[str, Any]:
                """Calls the OpenAI API and returns the API response."""

                api_key = os.environ.get("ANTHROPIC_API_KEY")
                client = AsyncAnthropic(api_key=api_key)

                response = await client.messages.create(
                    model=model_name,
                    max_tokens=self.max_tokens,
                    temperature=self.temperature,
                    system=system_prompt,
                    messages=[
                        #                        {"role": "system", "content": system_prompt},
                        {"role": "user", "content": user_prompt},
                    ],
                )
                return response.model_dump()

            @staticmethod
            def parse_response(raw_response: dict[str, Any]) -> str:
                """Parses the API response and returns the response text."""
                response = raw_response["content"][0]["text"]
                pattern = r"^```json(?:\\n|\n)(.+?)(?:\\n|\n)```$"
                match = re.match(pattern, response, re.DOTALL)
                if match:
                    return match.group(1)
                else:
                    return response

        LLM.__name__ = model_class_name

        return LLM



from __future__ import annotations
from collections import UserDict
import enum
import time

# from edsl.jobs.tasks.TaskStatusLogEntry import TaskStatusLogEntry


class TaskStatus(enum.Enum):
    "These are the possible states a task can be in."
    NOT_STARTED = enum.auto()
    WAITING_FOR_DEPENDENCIES = enum.auto()
    CANCELLED = enum.auto()
    PARENT_FAILED = enum.auto()
    WAITING_FOR_REQUEST_CAPACITY = enum.auto()
    WAITING_FOR_TOKEN_CAPACITY = enum.auto()
    API_CALL_IN_PROGRESS = enum.auto()
    SUCCESS = enum.auto()
    FAILED = enum.auto()


class TaskStatusLogEntry(UserDict):
    def __init__(self, log_time, value):
        self.data = {"log_time": log_time, "value": value}
        super().__init__(self.data)


class TaskStatusDescriptor:
    "The descriptor ensures that the task status is always an instance of the TaskStatus enum."

    def __init__(self):
        self._task_status = None

    def __get__(self, instance, owner):
        return self._task_status

    def __set__(self, instance, value):
        """Ensure that the value is an instance of TaskStatus."""
        if not isinstance(value, TaskStatus):
            raise ValueError("Value must be an instance of TaskStatus enum")
        t = time.monotonic()
        if hasattr(instance, "status_log"):
            instance.status_log.append(TaskStatusLogEntry(t, value))
        self._task_status = value

    def __delete__(self, instance):
        self._task_status = None


status_colors = {
    TaskStatus.NOT_STARTED: "grey",
    TaskStatus.WAITING_FOR_DEPENDENCIES: "orange",
    TaskStatus.WAITING_FOR_REQUEST_CAPACITY: "yellow",
    TaskStatus.WAITING_FOR_TOKEN_CAPACITY: "gold",
    TaskStatus.CANCELLED: "white",
    TaskStatus.PARENT_FAILED: "darkred",
    TaskStatus.FAILED: "red",
    TaskStatus.API_CALL_IN_PROGRESS: "blue",
    TaskStatus.SUCCESS: "green",
}


def get_enum_from_string(str_key):
    """Parse the string to extract the enum member name."""
    try:
        _, member_name = str_key.split(".")
        enum_member = getattr(TaskStatus, member_name)
        return enum_member
    except ValueError:
        return str_key


class InterviewTaskLogDict(UserDict):
    """A dictionary of TaskStatusLog objects.

    The key is the name of the task.
    """

    @property
    def min_time(self):
        return min([log.min_time for log in self.values()])

    @property
    def max_time(self):
        return max([log.max_time for log in self.values()])

    def status_matrix(self, num_periods):
        """Return a matrix of status values."""
        start_time = self.min_time
        end_time = self.max_time
        time_increment = (end_time - start_time) / num_periods
        status_matrix = {}
        time_periods = [start_time + i * time_increment for i in range(num_periods)]
        for task_name, log in self.items():
            status_matrix[task_name] = [log.status_at_time(t) for t in time_periods]
        return status_matrix

    def numerical_matrix(self, num_periods):
        """Return a numerical matrix of status values."""
        status_dicts = self.status_matrix(num_periods)

        num_cols = num_periods
        num_rows = len(status_dicts)
        matrix = [[0 for _ in range(num_cols)] for _ in range(num_rows)]

        for row_index, (task_name, status_list) in enumerate(status_dicts.items()):
            matrix[row_index] = [
                list(status_colors.keys()).index(status) for status in status_list
            ]

        index_to_names = {i: name for i, name in enumerate(status_dicts.keys())}
        return matrix, index_to_names

    def visualize(self, num_periods=10):
        """Visualize the status matrix with outlined squares."""
        import matplotlib.pyplot as plt
        from matplotlib.colors import ListedColormap
        import numpy as np
        from matplotlib.patches import Rectangle

        # Define your custom colormap
        custom_cmap = ListedColormap(list(status_colors.values()))

        # Generate the matrix
        matrix, index_to_names = self.numerical_matrix(num_periods)

        # Create the figure and axes
        plt.figure(figsize=(10, 5))
        ax = plt.gca()

        # Display the matrix and keep a reference to the imshow object
        im = ax.imshow(matrix, aspect="auto", cmap=custom_cmap)

        # Adding color bar, now correctly associating it with 'im'
        cbar = plt.colorbar(im, ticks=range(len(status_colors)), label="Task Status")

        cbar_labels = [status.name for status in status_colors.keys()]
        # breakpoint()
        cbar.set_ticklabels(cbar_labels)  # Setting the custom labels for the colorbar

        im.set_clim(
            -0.5, len(status_colors) - 0.5
        )  # Setting color limits directly on the imshow object

        # Outline each cell by drawing rectangles
        for (j, i), val in np.ndenumerate(matrix):
            ax.add_patch(
                Rectangle(
                    (i - 0.5, j - 0.5), 1, 1, fill=False, edgecolor="black", lw=0.5
                )
            )

        # Set custom y-axis ticks and labels
        yticks = list(index_to_names.keys())
        yticklabels = list(index_to_names.values())
        plt.yticks(ticks=yticks, labels=yticklabels)

        # Show the plot
        plt.show()


if __name__ == "__main__":
    pass



from collections import UserDict
from edsl.jobs.buckets.TokenBucket import TokenBucket
from edsl.jobs.buckets.ModelBuckets import ModelBuckets


class BucketCollection(UserDict):
    """A Jobs object will have a whole collection of model buckets, as multiple models could be used.

    The keys here are the models, and the values are the ModelBuckets objects.
    Models themselves are hashable, so this works.
    """

    def __init__(self, infinity_buckets=False):
        super().__init__()
        self.infinity_buckets = infinity_buckets

    def __repr__(self):
        return f"BucketCollection({self.data})"

    def add_model(self, model: "LanguageModel") -> None:
        """Adds a model to the bucket collection.

        This will create the token and request buckets for the model."""
        # compute the TPS and RPS from the model
        if not self.infinity_buckets:
            TPS = model.TPM / 60.0
            RPS = model.RPM / 60.0
        else:
            TPS = float("inf")
            RPS = float("inf")

        # create the buckets
        requests_bucket = TokenBucket(
            bucket_name=model.model,
            bucket_type="requests",
            capacity=RPS,
            refill_rate=RPS,
        )
        tokens_bucket = TokenBucket(
            bucket_name=model.model, bucket_type="tokens", capacity=TPS, refill_rate=TPS
        )
        model_buckets = ModelBuckets(requests_bucket, tokens_bucket)
        if model in self:
            # it if already exists, combine the buckets
            self[model] += model_buckets
        else:
            self[model] = model_buckets

    def visualize(self) -> dict:
        """Visualize the token and request buckets for each model."""
        plots = {}
        for model in self:
            plots[model] = self[model].visualize()
        return plots



"""A list of Agent objects.

Example usage:

.. code-block:: python

    al = AgentList([Agent.example(), Agent.example()])
    len(al)
    2
    
"""

from __future__ import annotations
import csv
import json
from collections import UserList
from typing import Any, List, Optional, Union
from rich import print_json
from rich.table import Table
from simpleeval import EvalWithCompoundTypes
from edsl.Base import Base
from edsl.utilities.decorators import add_edsl_version, remove_edsl_version


class AgentList(UserList, Base):
    """A list of Agents."""

    def __init__(self, data: Optional[list["Agent"]] = None):
        """Initialize a new AgentList.

        :param data: A list of Agents.
        """
        if data is not None:
            super().__init__(data)
        else:
            super().__init__()

    def shuffle(self, seed: Optional[str] = "edsl") -> AgentList:
        """Shuffle the AgentList.

        :param seed: The seed for the random number generator.
        """
        import random

        random.seed(seed)
        random.shuffle(self.data)
        return self

    def sample(self, n: int, seed="edsl") -> AgentList:
        """Return a random sample of agents.

        :param n: The number of agents to sample.
        :param seed: The seed for the random number generator.
        """
        import random

        random.seed(seed)
        return AgentList(random.sample(self.data, n))

    def rename(self, old_name, new_name):
        """Rename a trait in the AgentList.

        :param old_name: The old name of the trait.
        :param new_name: The new name of the trait.
        """
        for agent in self.data:
            agent.rename(old_name, new_name)
        return self

    def select(self, *traits) -> AgentList:
        """Selects agents with only the references traits.

        >>> from edsl.agents.Agent import Agent
        >>> al = AgentList([Agent(traits = {'a': 1, 'b': 1}), Agent(traits = {'a': 1, 'b': 2})])
        >>> al.select('a')
        AgentList([Agent(traits = {'a': 1}), Agent(traits = {'a': 1})])

        """

        if len(traits) == 1:
            traits_to_select = [list(traits)[0]]
        else:
            traits_to_select = list(traits)

        return AgentList([agent.select(*traits_to_select) for agent in self.data])

    def filter(self, expression: str) -> AgentList:
        """
        Filter a list of agents based on an expression.

        >>> from edsl.agents.Agent import Agent
        >>> al = AgentList([Agent(traits = {'a': 1, 'b': 1}), Agent(traits = {'a': 1, 'b': 2})])
        >>> al.filter("b == 2")
        AgentList([Agent(traits = {'a': 1, 'b': 2})])
        """

        def create_evaluator(agent: "Agent"):
            """Create an evaluator for the given result.
            The 'combined_dict' is a mapping of all values for that Result object.
            """
            return EvalWithCompoundTypes(names=agent.traits)

        try:
            # iterates through all the results and evaluates the expression
            new_data = [
                agent for agent in self.data if create_evaluator(agent).eval(expression)
            ]
        except Exception as e:
            print(f"Exception:{e}")
            raise Exception(f"Error in filter. Exception:{e}")

        return AgentList(new_data)

    @classmethod
    def from_csv(cls, file_path: str):
        """Load AgentList from a CSV file.

        >>> import csv
        >>> import os
        >>> with open('/tmp/agents.csv', 'w') as f:
        ...     writer = csv.writer(f)
        ...     _ = writer.writerow(['age', 'hair', 'height'])
        ...     _ = writer.writerow([22, 'brown', 5.5])
        >>> al = AgentList.from_csv('/tmp/agents.csv')
        >>> al
        AgentList([Agent(traits = {'age': '22', 'hair': 'brown', 'height': '5.5'})])
        >>> os.remove('/tmp/agents.csv')

        :param file_path: The path to the CSV file.
        """
        from edsl.agents.Agent import Agent

        agent_list = []
        with open(file_path, "r") as f:
            reader = csv.DictReader(f)
            for row in reader:
                agent_list.append(Agent(row))
        return cls(agent_list)

    def translate_traits(self, values_codebook: dict[str, str]):
        """Translate traits to a new codebook.

        :param codebook: The new codebook.
        """
        for agent in self.data:
            agent.translate_traits(codebook)
        return self

    def remove_trait(self, trait: str):
        """Remove traits from the AgentList.

        :param traits: The traits to remove.
        >>> from edsl.agents.Agent import Agent
        >>> al = AgentList([Agent({'age': 22, 'hair': 'brown', 'height': 5.5}), Agent({'age': 22, 'hair': 'brown', 'height': 5.5})])
        >>> al.remove_trait('age')
        AgentList([Agent(traits = {'hair': 'brown', 'height': 5.5}), Agent(traits = {'hair': 'brown', 'height': 5.5})])
        """
        for agent in self.data:
            _ = agent.remove_trait(trait)
        return self

    @staticmethod
    def get_codebook(file_path: str):
        """Return the codebook for a CSV file.

        :param file_path: The path to the CSV file.
        """
        with open(file_path, "r") as f:
            reader = csv.DictReader(f)
            return {field: None for field in reader.fieldnames}

    def __hash__(self) -> int:
        from edsl.utilities.utilities import dict_hash

        data = self.to_dict()
        # data['agent_list'] = sorted(data['agent_list'], key=lambda x: dict_hash(x)
        return dict_hash(self._to_dict(sorted=True))

    def _to_dict(self, sorted=False):
        if sorted:
            data = self.data[:]
            data.sort(key=lambda x: hash(x))
        else:
            data = self.data

        return {"agent_list": [agent.to_dict() for agent in data]}

    def __eq__(self, other: AgentList) -> bool:
        return self._to_dict(sorted=True) == other._to_dict(sorted=True)

    @add_edsl_version
    def to_dict(self):
        """Return dictionary of AgentList to serialization."""
        return self._to_dict()

    def __repr__(self):
        return f"AgentList({self.data})"

    def print(self, format: Optional[str] = None):
        """Print the AgentList."""
        print_json(json.dumps(self._to_dict()))

    def _repr_html_(self):
        """Return an HTML representation of the AgentList."""
        from edsl.utilities.utilities import data_to_html

        return data_to_html(self.to_dict()["agent_list"])

    def to_scenario_list(self) -> "ScenarioList":
        """Return a list of scenarios."""
        from edsl.scenarios.ScenarioList import ScenarioList
        from edsl.scenarios.Scenario import Scenario

        return ScenarioList([Scenario(agent.traits) for agent in self.data])

    @classmethod
    @remove_edsl_version
    def from_dict(cls, data: dict) -> "AgentList":
        """Deserialize the dictionary back to an AgentList object.

        :param: data: A dictionary representing an AgentList.
        >>> from edsl.agents.Agent import Agent
        >>> al = AgentList([Agent.example(), Agent.example()])
        >>> al2 = AgentList.from_dict(al.to_dict())
        >>> al2 == al
        True
        """
        from edsl.agents.Agent import Agent

        agents = [Agent.from_dict(agent_dict) for agent_dict in data["agent_list"]]
        return cls(agents)

    @classmethod
    def example(cls, randomize: bool = False) -> AgentList:
        """
        Returns an example AgentList instance.

        :param randomize: If True, uses Agent's randomize method.
        """
        from edsl.agents.Agent import Agent

        return cls([Agent.example(randomize), Agent.example(randomize)])

    @classmethod
    def from_list(self, trait_name: str, values: List[Any]):
        """Create an AgentList from a list of values.

        :param trait_name: The name of the trait.
        :param values: A list of values.
        """
        from edsl.agents.Agent import Agent

        return AgentList([Agent({trait_name: value}) for value in values])

    def __mul__(self, other: AgentList) -> AgentList:
        """Takes the cross product of two AgentLists."""
        from itertools import product

        new_sl = []
        for s1, s2 in list(product(self, other)):
            new_sl.append(s1 + s2)
        return AgentList(new_sl)

    def code(self, string=True) -> Union[str, list[str]]:
        """Return code to construct an AgentList.

        >>> al = AgentList.example()
        >>> print(al.code())
        from edsl.agents.Agent import Agent
        from edsl.agents.AgentList import AgentList
        agent_list = AgentList([Agent(traits = {'age': 22, 'hair': 'brown', 'height': 5.5}), Agent(traits = {'age': 22, 'hair': 'brown', 'height': 5.5})])
        """
        lines = [
            "from edsl.agents.Agent import Agent",
            "from edsl.agents.AgentList import AgentList",
        ]
        lines.append(f"agent_list = AgentList({self.data})")
        if string:
            return "\n".join(lines)
        return lines

    def rich_print(self) -> Table:
        """Display an object as a rich table."""
        table = Table(title="AgentList")
        table.add_column("Agents", style="bold")
        for agent in self.data:
            table.add_row(agent.rich_print())
        return table


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from edsl.utilities.gcp_bucket.cloud_storage import CloudStorageManager

manager = CloudStorageManager()

# Download Process
file_name = "GSS2022.dta"  # Name for the downloaded file
save_name = "test.dta"

manager.download_file(file_name, save_name)



# """This module contains the Result class, which captures the result of one interview."""
from __future__ import annotations
from collections import UserDict
from typing import Any, Type, Callable, Optional
from collections import UserDict
from edsl.Base import Base
from edsl.utilities.decorators import add_edsl_version, remove_edsl_version


class PromptDict(UserDict):
    """A dictionary that is used to store the prompt for a given result."""

    def rich_print(self):
        """Display an object as a table."""
        from rich.table import Table

        table = Table(title="")
        table.add_column("Attribute", style="bold")
        table.add_column("Value")

        to_display = self
        for attr_name, attr_value in to_display.items():
            table.add_row(attr_name, repr(attr_value))

        return table


def agent_namer_closure():
    """Return a function that can be used to name an agent."""
    agent_dict = {}

    def agent_namer(agent):
        """Return a name for an agent. If the agent has been named before, return the same name. Otherwise, return a new name."""
        nonlocal agent_dict
        agent_count = len(agent_dict)
        if id(agent) in agent_dict:
            return agent_dict[id(agent)]
        else:
            agent_dict[id(agent)] = f"Agent_{agent_count}"
            return agent_dict[id(agent)]

    return agent_namer


agent_namer = agent_namer_closure()


class Result(Base, UserDict):
    """
    This class captures the result of one interview.

    The answer dictionary has the structure:

    >>> import warnings
    >>> warnings.simplefilter("ignore", UserWarning)
    >>> Result.example().answer
    {'how_feeling': 'OK', 'how_feeling_comment': 'This is a real survey response from a human.', 'how_feeling_yesterday': 'Great', 'how_feeling_yesterday_comment': 'This is a real survey response from a human.'}

    Its main data is an Agent, a Scenario, a Model, an Iteration, and an Answer.
    These are stored both in the UserDict and as attributes.


    """

    def __init__(
        self,
        agent: "Agent",
        scenario: "Scenario",
        model: Type["LanguageModel"],
        iteration: int,
        answer: str,
        prompt: dict[str, str] = None,
        raw_model_response=None,
        survey: Optional["Survey"] = None,
        question_to_attributes: Optional[dict] = None,
    ):
        """Initialize a Result object.

        :param agent: The Agent object.
        :param scenario: The Scenario object.
        :param model: The LanguageModel object.
        :param iteration: The iteration number.
        :param answer: The answer string.
        :param prompt: A dictionary of prompts.
        :param raw_model_response: The raw model response.

        """
        if question_to_attributes is not None:
            question_to_attributes = question_to_attributes
        else:
            question_to_attributes = {}

        if survey is not None:
            question_to_attributes = {
                q.question_name: {
                    "question_text": q.question_text,
                    "question_type": q.question_type,
                    "question_options": (
                        None
                        if not hasattr(q, "question_options")
                        else q.question_options
                    ),
                }
                for q in survey.questions
            }

        data = {
            "agent": agent,
            "scenario": scenario,
            "model": model,
            "iteration": iteration,
            "answer": answer,
            "prompt": prompt or {},
            "raw_model_response": raw_model_response or {},
            "question_to_attributes": question_to_attributes,
        }
        super().__init__(**data)
        # but also store the data as attributes
        self.agent = agent
        self.scenario = scenario
        self.model = model
        self.iteration = iteration
        self.answer = answer
        self.prompt = prompt or {}
        self.raw_model_response = raw_model_response or {}
        self.survey = survey
        self.question_to_attributes = question_to_attributes

        self._combined_dict = None
        self._problem_keys = None

    ###############
    # Used in Results
    ###############
    @property
    def sub_dicts(self) -> dict[str, dict]:
        """Return a dictionary where keys are strings for each of the main class attributes/objects."""
        if self.agent.name is None:
            agent_name = agent_namer(self.agent)
        else:
            agent_name = self.agent.name

        comments_dict = {k: v for k, v in self.answer.items() if k.endswith("_comment")}
        question_text_dict = {}
        question_options_dict = {}
        question_type_dict = {}
        for key, _ in self.answer.items():
            if key in self.question_to_attributes:
                # You might be tempted to just use the naked key
                # but this is a bad idea because it pollutes the namespace
                question_text_dict[
                    key + "_question_text"
                ] = self.question_to_attributes[key]["question_text"]
                question_options_dict[
                    key + "_question_options"
                ] = self.question_to_attributes[key]["question_options"]
                question_type_dict[
                    key + "_question_type"
                ] = self.question_to_attributes[key]["question_type"]

        return {
            "agent": self.agent.traits
            | {"agent_name": agent_name}
            | {"agent_instruction": self.agent.instruction},
            "scenario": self.scenario,
            "model": self.model.parameters | {"model": self.model.model},
            "answer": self.answer,
            "prompt": self.prompt,
            "raw_model_response": self.raw_model_response,
            #            "iteration": {"iteration": self.iteration},
            "question_text": question_text_dict,
            "question_options": question_options_dict,
            "question_type": question_type_dict,
            "comment": comments_dict,
        }

    def check_expression(self, expression) -> None:
        for key in self.problem_keys:
            if key in expression and not key + "." in expression:
                raise ValueError(
                    f"Key by iself {key} is problematic. Use the full key {key + '.' + key} name instead."
                )
        return None

    def code(self):
        """Return a string of code that can be used to recreate the Result object."""
        raise NotImplementedError

    @property
    def problem_keys(self):
        """Return a list of keys that are problematic."""
        return self._problem_keys

    def _compute_combined_dict_and_problem_keys(self) -> None:
        combined = {}
        problem_keys = []
        for key, sub_dict in self.sub_dicts.items():
            combined.update(sub_dict)
            # in some cases, the sub_dict might have keys that conflict with the main dict
            if key in combined:
                # The key is already in the combined dict
                problem_keys = problem_keys + [key]

            combined.update({key: sub_dict})
            # I *think* this allows us to do do things like "answer.how_feelling" i.e., that the evaluator can use
            # dot notation to access the subdicts.
        self._combined_dict = combined
        self._problem_keys = problem_keys

    @property
    def combined_dict(self) -> dict[str, Any]:
        """Return a dictionary that includes all sub_dicts, but also puts the key-value pairs in each sub_dict as a key_value pair in the combined dictionary.

        >>> r = Result.example()
        >>> r.combined_dict['how_feeling']
        'OK'
        """
        if self._combined_dict is None or self._problem_keys is None:
            self._compute_combined_dict_and_problem_keys()
        return self._combined_dict

    @property
    def problem_keys(self):
        """Return a list of keys that are problematic."""
        if self._combined_dict is None or self._problem_keys is None:
            self._compute_combined_dict_and_problem_keys()
        return self._problem_keys

    def get_value(self, data_type: str, key: str) -> Any:
        """Return the value for a given data type and key.

        >>> r = Result.example()
        >>> r.get_value("answer", "how_feeling")
        'OK'

        - data types can be "agent", "scenario", "model", or "answer"
        - keys are relevant attributes of the Objects the data types represent
        """
        return self.sub_dicts[data_type][key]

    @property
    def key_to_data_type(self) -> dict[str, str]:
        """Return a dictionary where keys are object attributes and values are the data type (object) that the attribute is associated with.

        >>> r = Result.example()
        >>> r.key_to_data_type["how_feeling"]
        'answer'

        """
        d = {}
        data_types = self.sub_dicts.keys()
        for data_type in data_types:
            for key in self.sub_dicts[data_type]:
                d[key] = data_type
        return d

    def rows(self, index) -> tuple[int, str, str, str]:
        """Return a generator of rows for the Result object."""
        for data_type, subdict in self.sub_dicts.items():
            for key, value in subdict.items():
                yield (index, data_type, key, str(value))

    ###############
    # Useful
    ###############
    def copy(self) -> Result:
        """Return a copy of the Result object."""
        return Result.from_dict(self.to_dict())

    def __eq__(self, other) -> bool:
        """Return True if the Result object is equal to another Result object.

        >>> r = Result.example()
        >>> r == r
        True

        """
        return self.to_dict() == other.to_dict()

    ###############
    # Serialization
    ###############
    def _to_dict(self) -> dict[str, Any]:
        """Return a dictionary representation of the Result object."""
        d = {}
        for key, value in self.items():
            if hasattr(value, "to_dict"):
                d[key] = value.to_dict()
            else:
                d[key] = value
            if key == "prompt":
                new_prompt_dict = {}
                for prompt_name, prompt_obj in value.items():
                    new_prompt_dict[prompt_name] = (
                        prompt_obj
                        if not hasattr(prompt_obj, "to_dict")
                        else prompt_obj.to_dict()
                    )
                d[key] = new_prompt_dict
        return d

    @add_edsl_version
    def to_dict(self) -> dict[str, Any]:
        """Return a dictionary representation of the Result object.

        >>> r = Result.example()
        >>> r.to_dict()['scenario']
        {'period': 'morning', 'edsl_version': '...', 'edsl_class_name': 'Scenario'}
        """
        return self._to_dict()

    def __hash__(self):
        """Return a hash of the Result object."""
        from edsl.utilities.utilities import dict_hash

        return dict_hash(self._to_dict())

    @classmethod
    @remove_edsl_version
    def from_dict(self, json_dict: dict) -> Result:
        """Return a Result object from a dictionary representation."""

        from edsl import Agent
        from edsl import Scenario
        from edsl.language_models.LanguageModel import LanguageModel
        from edsl.prompts.Prompt import Prompt

        prompt_data = json_dict.get("prompt", {})
        prompt_d = {}
        for prompt_name, prompt_obj in prompt_data.items():
            prompt_d[prompt_name] = Prompt.from_dict(prompt_obj)

        result = Result(
            agent=Agent.from_dict(json_dict["agent"]),
            scenario=Scenario.from_dict(json_dict["scenario"]),
            model=LanguageModel.from_dict(json_dict["model"]),
            iteration=json_dict["iteration"],
            answer=json_dict["answer"],
            prompt=prompt_d,  # json_dict["prompt"],
            raw_model_response=json_dict.get(
                "raw_model_response", {"raw_model_response": "No raw model response"}
            ),
            question_to_attributes=json_dict.get("question_to_attributes", None),
        )
        return result

    def rich_print(self) -> None:
        """Display an object as a table."""
        # from edsl.utilities import print_dict_with_rich
        from rich import print
        from rich.table import Table

        table = Table(title="Result")
        table.add_column("Attribute", style="bold")
        table.add_column("Value")

        to_display = self.__dict__.copy()
        data = to_display.pop("data", None)
        for attr_name, attr_value in to_display.items():
            if hasattr(attr_value, "rich_print"):
                table.add_row(attr_name, attr_value.rich_print())
            elif isinstance(attr_value, dict):
                a = PromptDict(attr_value)
                table.add_row(attr_name, a.rich_print())
            else:
                table.add_row(attr_name, repr(attr_value))
        return table

    def __repr__(self):
        """Return a string representation of the Result object."""
        return f"Result(agent={repr(self.agent)}, scenario={repr(self.scenario)}, model={repr(self.model)}, iteration={self.iteration}, answer={repr(self.answer)}, prompt={repr(self.prompt)})"

    @classmethod
    def example(cls):
        """Return an example Result object."""
        from edsl.results.Results import Results

        return Results.example()[0]

    def score(self, scoring_function: Callable) -> Any:
        """Score the result using a passed-in scoring function.

        >>> def f(status): return 1 if status == 'Joyful' else 0
        >>> Result.example().score(f)
        1
        """
        import inspect

        signature = inspect.signature(scoring_function)
        params = {}
        for k, v in signature.parameters.items():
            if k in self.combined_dict:
                params[k] = self.combined_dict[k]
            elif v.default is not v.empty:
                params[k] = v.default
            else:
                raise ValueError(f"Parameter {k} not found in Result object")
        return scoring_function(**params)


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from edsl.jobs.tasks.task_status_enum import TaskStatus
from typing import List, Optional
from io import BytesIO
import base64


class TaskHistory:
    def __init__(self, interviews: List["Interview"], include_traceback=False):
        """
        The structure of a TaskHistory exception

        [Interview.exceptions, Interview.exceptions, Interview.exceptions, ...]

        """

        self.total_interviews = interviews
        self.include_traceback = include_traceback

        self._interviews = {index: i for index, i in enumerate(self.total_interviews)}

    @property
    def exceptions(self):
        return [i.exceptions for k, i in self._interviews.items() if i.exceptions != {}]

    @property
    def indices(self):
        return [k for k, i in self._interviews.items() if i.exceptions != {}]

    def __repr__(self):
        """Return a string representation of the TaskHistory."""
        return f"TaskHistory(interviews={self.total_interviews})."

    def to_dict(self):
        """Return the TaskHistory as a dictionary."""
        return {
            "exceptions": [
                e.to_dict(include_traceback=self.include_traceback)
                for e in self.exceptions
            ],
            "indices": self.indices,
        }

    @property
    def has_exceptions(self) -> bool:
        """Return True if there are any exceptions."""
        return len(self.exceptions) > 0

    def _repr_html_(self):
        """Return an HTML representation of the TaskHistory."""
        from edsl.utilities.utilities import data_to_html

        newdata = self.to_dict()["exceptions"]
        return data_to_html(newdata, replace_new_lines=True)

    def show_exceptions(self, tracebacks=False):
        """Print the exceptions."""
        for index in self.indices:
            self.total_interviews[index].exceptions.print(tracebacks)

    def get_updates(self):
        """Return a list of all the updates."""
        updates = []
        for interview in self.total_interviews:
            for question_name, logs in interview.task_status_logs.items():
                updates.append(logs)
        return updates

    def print(self):
        from rich import print

        print(self.get_updates())

    def plot_completion_times(self):
        """Plot the completion times for each task."""
        import matplotlib.pyplot as plt

        updates = self.get_updates()

        elapsed = [update.max_time - update.min_time for update in updates]
        for i, update in enumerate(updates):
            if update[-1]["value"] != TaskStatus.SUCCESS:
                elapsed[i] = 0
        x = range(len(elapsed))
        y = elapsed

        plt.bar(x, y)
        plt.title("Per-interview completion times")
        plt.xlabel("Task")
        plt.ylabel("Time (seconds)")
        plt.show()

    def plotting_data(self, num_periods=100):
        updates = self.get_updates()

        min_t = min([update.min_time for update in updates])
        max_t = max([update.max_time for update in updates])
        delta_t = (max_t - min_t) / (num_periods * 1.0)
        time_periods = [min_t + delta_t * i for i in range(num_periods)]

        def counts(t):
            d = {}
            for update in updates:
                status = update.status_at_time(t)
                if status in d:
                    d[status] += 1
                else:
                    d[status] = 1
            return d

        status_counts = [counts(t) for t in time_periods]

        new_counts = []
        for status_count in status_counts:
            d = {task_status: 0 for task_status in TaskStatus}
            d.update(status_count)
            new_counts.append(d)

        return new_counts

    def plot(self, num_periods=100, get_embedded_html=False):
        """Plot the number of tasks in each state over time."""
        new_counts = self.plotting_data(num_periods)
        max_count = max([max(entry.values()) for entry in new_counts])

        rows = int(len(TaskStatus) ** 0.5) + 1
        cols = (len(TaskStatus) + rows - 1) // rows  # Ensure all plots fit

        import matplotlib.pyplot as plt

        fig, axes = plt.subplots(rows, cols, figsize=(15, 10))
        axes = axes.flatten()  # Flatten in case of a single row/column

        for i, status in enumerate(TaskStatus):
            ax = axes[i]
            x = range(len(new_counts))
            y = [
                item.get(status, 0) for item in new_counts
            ]  # Use .get() to handle missing keys safely
            ax.plot(x, y, marker="o", linestyle="-")
            ax.set_title(status.name)
            ax.set_xlabel("Time Periods")
            ax.set_ylabel("Count")
            ax.grid(True)
            ax.set_ylim(0, max_count)

        # Hide any unused subplots
        for ax in axes[len(TaskStatus) :]:
            ax.axis("off")

        plt.tight_layout()

        if get_embedded_html:
            buffer = BytesIO()
            fig.savefig(buffer, format="png")
            plt.close(fig)
            buffer.seek(0)

            # Encode plot to base64 string
            img_data = base64.b64encode(buffer.getvalue()).decode("utf-8")
            buffer.close()
            return f'<img src="data:image/png;base64,{img_data}" alt="Plot">'
        else:
            plt.show()

    def css(self):
        return """
        body {
        font-family: Arial, sans-serif;
        line-height: 1.6;
        background-color: #f9f9f9;
        color: #333;
        margin: 20px;
        }

        .interview {
        font-size: 1.5em;
        margin-bottom: 10px;
        padding: 10px;
        background-color: #e3f2fd;
        border-left: 5px solid #2196f3;
        }

        .question {
        font-size: 1.2em;
        margin-bottom: 10px;
        padding: 10px;
        background-color: #fff9c4;
        border-left: 5px solid #ffeb3b;
        }

        .exception-detail {
        margin-bottom: 10px;
        padding: 10px;
        background-color: #ffebee;
        border-left: 5px solid #f44336;
        }

        .question-detail {
           border: 3px solid black; /* Adjust the thickness by changing the number */
            padding: 10px; /* Optional: Adds some padding inside the border */
        }

        .exception-detail div {
        margin-bottom: 5px;
        }

        .exception-exception {
        font-weight: bold;
        color: #d32f2f;
        }

        .exception-time,
        .exception-traceback {
        font-style: italic;
        color: #555;
        }
        """

    def html(
        self,
        filename: Optional[str] = None,
        return_link=False,
        css=None,
        cta="Open Report in New Tab",
    ):
        """Return an HTML report."""

        from IPython.display import display, HTML
        import tempfile
        import os
        from edsl.utilities.utilities import is_notebook
        from jinja2 import Template

        performance_plot_html = self.plot(num_periods=100, get_embedded_html=True)

        if css is None:
            css = self.css()

        template = Template(
            """
        <!DOCTYPE html>
        <html lang="en">
        <head>
        <meta charset="UTF-8">
        <meta name="viewport" content="width=device-width, initial-scale=1.0">
        <title>Exception Details</title>
        <style>
        {{ css }}
        </style>
        </head>
        <body>
            {% for index, interview in interviews.items() %}
                {% if interview.exceptions != {} %}
                   <div class="interview">Interview: {{ index }} </div>
                    <h1>Failing questions</h1>
                {% endif %}
                {% for question, exceptions in interview.exceptions.items() %}
                    <div class="question">question_name: {{ question }}</div>

                    <h2>Question</h2>
                    <div class="question-detail"> 
                            {{ interview.survey.get_question(question).html() }}
                    </div>        

                    <h2>Scenario</h2>                            
                    <div class="scenario"> 
                            {{ interview.scenario._repr_html_() }}
                    </div>        

                    <h2>Agent</h2>
                    <div class="agent">
                            {{ interview.agent._repr_html_() }}
                    </div>

                    <h2>Model</h2>
                    <div class="model">
                            {{ interview.model._repr_html_() }}
                    </div>
                            
                    <h2>Exception details</h2>

                    {% for exception_message in exceptions %}
                        <div class="exception-detail">
                            <div class="exception-exception">Exception: {{ exception_message.exception }}</div>
                            <div class="exception-time">Time: {{ exception_message.time }}</div>
                            <div class="exception-traceback">Traceback: <pre>{{ exception_message.traceback }} </pre></div>
                        </div>
                    {% endfor %}            
                {% endfor %}            
            {% endfor %}
                            
        <h1>Performance Plot</h1>
        {{ performance_plot_html }}
        </body>
        </html>
        """
        )

        # Render the template with data
        output = template.render(
            interviews=self._interviews,
            css=css,
            performance_plot_html=performance_plot_html,
        )

        # Save the rendered output to a file
        with open("output.html", "w") as f:
            f.write(output)

        if css is None:
            css = self.css()

        if filename is None:
            current_directory = os.getcwd()
            filename = tempfile.NamedTemporaryFile(
                "w", delete=False, suffix=".html", dir=current_directory
            ).name

        with open(filename, "w") as f:
            with open(filename, "w") as f:
                # f.write(html_header)
                # f.write(self._repr_html_())
                f.write(output)
                # f.write(html_footer)

        if is_notebook():
            import html

            html_url = f"/files/{filename}"
            html_link = f'<a href="{html_url}" target="_blank">{cta}</a>'
            display(HTML(html_link))
            escaped_output = html.escape(output)
            iframe = f""""
            <iframe srcdoc="{ escaped_output }" style="width: 800px; height: 600px;"></iframe>
            """
            display(HTML(iframe))
            # display(HTML(output))
        else:
            print(f"Exception report saved to {filename}")
            import webbrowser
            import os

            webbrowser.open(f"file://{os.path.abspath(filename)}")

        if return_link:
            return filename



import json
import numpy as np
from sklearn.manifold import TSNE
from IPython.display import display_html


def compute_tsne(embeddings, labels):
    embeddings_np = np.array(embeddings)
    tsne = TSNE(n_components=2, random_state=42)
    tsne_results = tsne.fit_transform(embeddings_np)
    data = [
        {
            "x": float(tsne_results[i, 0]),
            "y": float(tsne_results[i, 1]),
            "label": labels[i],
        }
        for i in range(len(labels))
    ]
    return data


def plot_tsne_in_notebook(embeddings, labels):
    # Compute t-SNE
    data = compute_tsne(embeddings, labels)

    # Convert data to JSON
    data_json = json.dumps(data)

    # HTML content with embedded data
    html_content = f"""
    <!DOCTYPE html>
    <html lang="en">
    <head>
        <meta charset="UTF-8">
        <title>t-SNE Plot with D3.js</title>
        <script src="https://d3js.org/d3.v6.min.js"></script>
        <style>
            .tooltip {{
                position: absolute;
                text-align: center;
                width: 80px;
                height: 28px;
                padding: 2px;
                font: 12px sans-serif;
                background: lightsteelblue;
                border: 0px;
                border-radius: 8px;
                pointer-events: none;
            }}
        </style>
    </head>
    <body>
        <svg width="600" height="600"></svg>

        <script>
            // Embedded data
            const data = {data_json};

            const svg = d3.select("svg"),
                width = +svg.attr("width"),
                height = +svg.attr("height");

            // Set up scales
            const x = d3.scaleLinear()
                .domain(d3.extent(data, d => d.x))
                .range([0, width]);
            
            const y = d3.scaleLinear()
                .domain(d3.extent(data, d => d.y))
                .range([height, 0]);

            // Create tooltip
            const tooltip = d3.select("body").append("div")
                .attr("class", "tooltip")
                .style("opacity", 0);

            // Create circles for each point
            svg.selectAll("circle")
                .data(data)
                .enter().append("circle")
                .attr("cx", d => x(d.x))
                .attr("cy", d => y(d.y))
                .attr("r", 5)
                .style("fill", "steelblue")
                .on("mouseover", function(event, d) {{
                    tooltip.transition()
                        .duration(200)
                        .style("opacity", .9);
                    tooltip.html(d.label)
                        .style("left", (event.pageX + 5) + "px")
                        .style("top", (event.pageY - 28) + "px");
                }})
                .on("mouseout", function(d) {{
                    tooltip.transition()
                        .duration(500)
                        .style("opacity", 0);
                }});
        </script>
    </body>
    </html>
    """

    # Write HTML content to a temporary file
    html_file = "tsne_plot.html"
    with open(html_file, "w") as file:
        file.write(html_content)

    # Display the HTML content in an iframe within a Jupyter notebook
    display_html(
        f'<iframe src="{html_file}" width="600" height="600"></iframe>', raw=True
    )


# Example usage
if __name__ == "__main__":
    embeddings = [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6], [0.7, 0.8, 0.9]]
    labels = ["String 1", "String 2", "String 3"]
    plot_tsne_in_notebook(embeddings, labels)



"""This module provides a Config class that loads environment variables from a .env file and sets them as class attributes."""

import os
from edsl.exceptions import (
    InvalidEnvironmentVariableError,
    MissingEnvironmentVariableError,
)
from dotenv import load_dotenv, find_dotenv

# valid values for EDSL_RUN_MODE
EDSL_RUN_MODES = ["development", "development-testrun", "production"]

# `default` is used to impute values only in "production" mode
# `info` gives a brief description of the env var
CONFIG_MAP = {
    "EDSL_RUN_MODE": {
        "default": "production",
        "info": "This env var determines the run mode of the application.",
    },
    "EDSL_DATABASE_PATH": {
        "default": f"sqlite:///{os.path.join(os.getcwd(), '.edsl_cache/data.db')}",
        "info": "This env var determines the path to the cache file.",
    },
    "EDSL_LOGGING_PATH": {
        "default": f"{os.path.join(os.getcwd(), 'interview.log')}",
        "info": "This env var determines the path to the log file.",
    },
    "EDSL_API_TIMEOUT": {
        "default": "60",
        "info": "This env var determines the maximum number of seconds to wait for an API call to return.",
    },
    "EDSL_BACKOFF_START_SEC": {
        "default": "1",
        "info": "This env var determines the number of seconds to wait before retrying a failed API call.",
    },
    "EDSL_MAX_BACKOFF_SEC": {
        "default": "60",
        "info": "This env var determines the maximum number of seconds to wait before retrying a failed API call.",
    },
    "EDSL_MAX_ATTEMPTS": {
        "default": "5",
        "info": "This env var determines the maximum number of times to retry a failed API call.",
    },
    "EXPECTED_PARROT_URL": {
        "default": "https://www.expectedparrot.com",
        "info": "This env var holds the URL of the Expected Parrot API.",
    },
    # "EXPECTED_PARROT_API_KEY": {
    #     "default": None,
    #     "info": "This env var holds your Expected Parrot API key (https://www.expectedparrot.com/).",
    # },
    # "OPENAI_API_KEY": {
    #     "default": None,
    #     "info": "This env var holds your OpenAI API key (https://platform.openai.com/api-keys).",
    # },
    # "DEEP_INFRA_API_KEY": {
    #     "default": None,
    #     "info": "This env var holds your DeepInfra API key (https://deepinfra.com/).",
    # },
    # "GOOGLE_API_KEY": {
    #     "default": None,
    #     "info": "This env var holds your Google API key (https://console.cloud.google.com/apis/credentials).",
    # },
    # "ANTHROPIC_API_KEY": {
    #     "default": None,
    #     "info": "This env var holds your Anthropic API key (https://www.anthropic.com/).",
    # },
}


class Config:
    """A class that loads environment variables from a .env file and sets them as class attributes."""

    def __init__(self):
        """Initialize the Config class."""
        self._set_run_mode()
        self._load_dotenv()
        self._set_env_vars()

    def _set_run_mode(self) -> None:
        """
        Checks the validity and sets EDSL_RUN_MODE.
        """
        run_mode = os.getenv("EDSL_RUN_MODE")
        default = CONFIG_MAP.get("EDSL_RUN_MODE").get("default")
        if run_mode is None:
            run_mode = default
        if run_mode not in EDSL_RUN_MODES:
            raise InvalidEnvironmentVariableError(
                f"Value `{run_mode}` is not allowed for EDSL_RUN_MODE."
            )
        self.EDSL_RUN_MODE = run_mode

    def _load_dotenv(self) -> None:
        """
        Loads the .env
        - Overrides existing env vars unless EDSL_RUN_MODE=="development-testrun"
        """

        override = True
        if self.EDSL_RUN_MODE == "development-testrun":
            override = False
        _ = load_dotenv(dotenv_path=find_dotenv(usecwd=True), override=override)

    def _set_env_vars(self) -> None:
        """
        Sets env vars as Config class attributes.
        - If an env var is not set and has a default value in the CONFIG_MAP, sets it to the default value.
        """
        # for each env var in the CONFIG_MAP
        for env_var, config in CONFIG_MAP.items():
            if env_var == "EDSL_RUN_MODE":
                continue  # we've set it already in _set_run_mode
            value = os.getenv(env_var)
            default_value = config.get("default")
            # if the env var is set, set it as a CONFIG attribute
            if value:
                setattr(self, env_var, value)
            # otherwise, if EDSL_RUN_MODE == "production" set it to its default value
            elif self.EDSL_RUN_MODE == "production":
                setattr(self, env_var, default_value)

    def get(self, env_var: str) -> str:
        """
        Returns the value of an environment variable.
        """
        if env_var not in CONFIG_MAP:
            raise InvalidEnvironmentVariableError(f"{env_var} is not a valid env var. ")
        elif env_var not in self.__dict__:
            info = CONFIG_MAP[env_var].get("info")
            raise MissingEnvironmentVariableError(f"{env_var} is not set. {info}")
        return self.__dict__.get(env_var)

    def show(self) -> str:
        """Print the currently set environment vars."""
        max_env_var_length = max(len(env_var) for env_var in self.__dict__)
        print("Here are the current configuration settings:")
        for env_var, value in self.__dict__.items():
            print(f"{env_var:<{max_env_var_length}} : {value}")


# Note: Python modules are singletons. As such, once this module is imported
# the same instance of it is reused across the application.
CONFIG = Config()



import hashlib
import time
from typing import Any, Dict, List, Optional


class ProofOfWork:
    def __init__(
        self,
        input_data: Optional[Any] = None,
        proof: Optional[Dict[int, List[Dict[str, Any]]]] = None,
    ):
        self.input_data = input_data
        self.proof = proof or {}

    def add_input_data(self, input_data: Any) -> None:
        self.input_data = input_data

    def to_dict(self) -> Dict[str, Any]:
        return {
            "input_data": self.input_data,
            "proof": self.proof,
        }

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "ProofOfWork":
        return cls(data["input_data"], data["proof"])

    def __repr__(self) -> str:
        return f"ProofOfWork(input_data={self.input_data}, proof={self.proof})"

    def to_hash(self, nonce: int) -> str:
        """
        Hash the input data combined with the nonce.

        Returns:
        str: The resulting hash.
        """
        hash_input = self.input_data + str(nonce)
        return hashlib.md5(hash_input.encode()).hexdigest()

    def verify_work(self) -> bool:
        for difficulty in self.proof:
            for proof in self.proof[difficulty]:
                nonce = proof["nonce"]
                hash_result = self.to_hash(nonce)
                prefix = "0" * difficulty
                if not hash_result.startswith(prefix):
                    return False
                if hash_result != proof["hash"]:
                    return False
        return True

    def add_proof(self, difficulty: int, starting_nonce: Optional[int] = None) -> None:
        """
        Find a nonce that results in a hash with `difficulty` number of leading zeros.

        Returns:
        int, str: The nonce that solves the proof of work and the resulting hash.
        """
        # Convert the difficulty into a string of zeros for comparison
        prefix = "0" * difficulty
        if not starting_nonce:
            import random

            starting_nonce = random.randint(0, 1000000)
        nonce = starting_nonce
        start = time.time()
        while True:
            # Combine the input data with the nonce and hash it
            hash_result = self.to_hash(nonce)

            # Check if the hash meets the difficulty requirement
            if hash_result.startswith(prefix):
                cycles = nonce - starting_nonce
                end = time.time()
                if difficulty in self.proof:
                    self.proof[difficulty].append(
                        {
                            "nonce": nonce,
                            "hash": hash_result,
                            "time": end - start,
                            "cycles": cycles,
                        }
                    )
                else:
                    self.proof[difficulty] = [
                        {
                            "nonce": nonce,
                            "hash": hash_result,
                            "time": end - start,
                            "cycles": cycles,
                        }
                    ]
                return

            nonce += 1


if __name__ == "__main__":
    from edsl.study import ProofOfWork

    p = ProofOfWork("hello world")
    p.add_proof(3)
    print(p)
    p.add_proof(6)
    print(p)

    # Takes about a minute to run
    p.add_proof(7)
    print(p)

    ok = p.verify_work()
    print(ok)
